from __future__ import annotations

from collections.abc import Container as CollectionsContainer
from collections.abc import Sequence as CollectionsSequence
from itertools import chain, product
from typing import (
    TYPE_CHECKING,
    Any,
    Collection,
    Container,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    overload,
)

from pharmpy.internals.expr.eval import eval_expr
from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.internals.expr.subs import subs, xreplace_dict
from pharmpy.internals.immutable import Immutable
from pharmpy.internals.math import cov2corr, is_positive_semidefinite, nearest_positive_semidefinite

from .distributions.numeric import NumericDistribution
from .distributions.symbolic import Distribution, JointNormalDistribution, NormalDistribution

if TYPE_CHECKING:
    import numpy as np
    import symengine
    import sympy
    import sympy.stats as sympy_stats
else:
    from pharmpy.deps import numpy as np
    from pharmpy.deps import symengine, sympy, sympy_stats


def _create_rng(seed: Optional[Union[int, np.random.Generator]] = None) -> np.random.Generator:
    """Create a new random number generator"""
    if isinstance(seed, np.random.Generator):
        return seed
    else:
        return np.random.default_rng(seed)


class VariabilityLevel(Immutable):
    """A variability level

    Parameters
    ----------
    name : str
        A unique identifying name
    reference : bool
        Is this the reference level? Normally IIV would be the reference level
    group : str
        Name of data column to group this level. None for no grouping (default)
    """

    def __init__(self, name: str, reference: bool = False, group: Optional[str] = None):
        self._name = name
        self._reference = reference
        self._group = group

    @classmethod
    def create(cls, name: str, reference: bool = False, group: Optional[str] = None):
        return VariabilityLevel(name, bool(reference), group)

    def replace(self, **kwargs) -> VariabilityLevel:
        name = kwargs.get('name', self._name)
        reference = kwargs.get('reference', self._reference)
        group = kwargs.get('group', self._group)
        return VariabilityLevel(name, bool(reference), group)

    def __eq__(self, other):
        return (
            isinstance(other, VariabilityLevel)
            and self._name == other._name
            and self._reference == other._reference
            and self._group == other._group
        )

    def __hash__(self):
        return hash((self._name, self._reference, self._group))

    def to_dict(self) -> Dict[str, Any]:
        return {'name': self._name, 'reference': self._reference, 'group': self._group}

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> VariabilityLevel:
        return cls(**d)

    @property
    def name(self) -> str:
        """Name of the variability level"""
        return self._name

    @property
    def reference(self) -> bool:
        """Is this the reference level"""
        return self._reference

    @property
    def group(self) -> Optional[str]:
        """Group variable for variability level"""
        return self._group

    def __repr__(self):
        return f"VariabilityLevel({self._name}, reference={self._reference}, group={self._group})"


class VariabilityHierarchy(Immutable):
    """Description of a variability hierarchy"""

    def __init__(self, levels: Tuple[VariabilityLevel, ...] = ()):
        self._levels = levels

    @classmethod
    def create(
        cls, levels: Optional[Union[Sequence[VariabilityLevel], VariabilityHierarchy]] = None
    ):
        if levels is None:
            levels = ()
        elif isinstance(levels, VariabilityHierarchy):
            return levels
        else:
            found_ref = False
            for level in levels:
                if not isinstance(level, VariabilityLevel):
                    raise ValueError("Can only add VariabilityLevel to VariabilityHierarchy")
                if level.reference:
                    if found_ref:
                        raise ValueError("A VariabilityHierarchy can only have one reference level")
                    else:
                        found_ref = True
            if not found_ref:
                raise ValueError("A VariabilityHierarchy must have a reference level")
            levels = tuple(levels)
        return VariabilityHierarchy(levels)

    def replace(self, **kwargs) -> VariabilityHierarchy:
        """Replace properties and create a new VariabilityHierarchy object"""
        levels = kwargs.get('levels', self._levels)
        new = VariabilityHierarchy.create(levels)
        return new

    def __eq__(self, other):
        if not isinstance(other, VariabilityHierarchy):
            return False

        if len(self._levels) != len(other._levels):
            return False
        else:
            for l1, l2 in zip(self._levels, other._levels):
                if l1 != l2:
                    return False
            return True

    def __hash__(self):
        return hash(self._levels)

    def to_dict(self) -> Dict[str, Any]:
        levels = tuple(level.to_dict() for level in self._levels)
        return {'levels': levels}

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> VariabilityHierarchy:
        levels = tuple(VariabilityLevel.from_dict(vl) for vl in d['levels'])
        return cls(levels=levels)

    def _lookup(self, ind: Union[int, str, VariabilityLevel]) -> VariabilityLevel:
        # Lookup one index
        if isinstance(ind, int):
            # Index on numeric level for ints
            i = self._find_reference()
            return self._levels[ind - i]
        elif isinstance(ind, str):
            for varlev in self._levels:
                if varlev.name == ind:
                    return varlev
        elif isinstance(ind, VariabilityLevel):
            for varlev in self._levels:
                if varlev.name == ind.name:
                    return varlev

        raise KeyError(f'Could not find level {ind} in VariabilityHierarchy')

    @overload
    def __getitem__(self, ind: Union[Sequence, VariabilityHierarchy]) -> VariabilityHierarchy:
        ...

    @overload
    def __getitem__(self, ind: Union[int, str, VariabilityLevel]) -> VariabilityLevel:
        ...

    def __getitem__(self, ind):
        if isinstance(ind, VariabilityHierarchy):
            levels = [level.name for level in ind._levels]
        elif not isinstance(ind, str) and isinstance(ind, CollectionsSequence):
            levels = ind
        else:
            return self._lookup(ind)
        new = [self._lookup(level) for level in levels]
        return VariabilityHierarchy.create(new)

    def __add__(self, other: VariabilityLevel) -> VariabilityHierarchy:
        if isinstance(other, VariabilityLevel):
            levels = (other,)
        else:
            raise ValueError(f"Cannot add {other} to VariabilityHierarchy")
        new = VariabilityHierarchy.create(self._levels + levels)
        return new

    def __radd__(self, other: VariabilityLevel) -> VariabilityHierarchy:
        if isinstance(other, VariabilityLevel):
            return VariabilityHierarchy.create((other,) + self._levels)
        else:
            raise ValueError(f"Cannot add {other} to VariabilityLevel")

    @property
    def names(self) -> List[str]:
        """Names of all variability levels"""
        return [varlev.name for varlev in self._levels]

    def _find_reference(self) -> int:
        # Find numerical level of first level
        # No error checking since having a reference level is an invariant
        return next(  # pragma: no cover
            (-i for i, level in enumerate(self._levels) if level.reference)
        )

    @property
    def levels(self) -> Dict[str, int]:
        """Dictionary of variability level name to numerical level"""
        ind = self._find_reference()
        d = {}
        for level in self._levels:
            d[level.name] = ind
            ind += 1
        return d

    def __len__(self):
        return len(self._levels)

    def __contains__(self, value):
        return value in self.names


class RandomVariables(CollectionsSequence, Immutable):
    """A collection of distributions of random variables

    This class provides a container for random variables that preserves their order
    and acts list-like while also allowing for indexing on names.

    Each RandomVariables object has two VariabilityHierarchies that describes the
    allowed variability levels for the contined random variables. One hierarchy
    for residual error variability (epsilons) and one for parameter variability (etas).
    By default the eta hierarchy has the two levels IIV and IOV and the epsilon
    hierarchy has one single level.

    Parameters
    ----------
    rvs : list
        A list of RandomVariable to add. Default is to create an empty RandomVariables.

    Examples
    --------
    >>> from pharmpy.model import RandomVariables, NormalDistribution, Parameter
    >>> omega = Parameter("OMEGA_CL", 0.1)
    >>> dist = NormalDistribution.create("IIV_CL", "iiv", 0, omega.symbol)
    >>> rvs = RandomVariables.create([dist])
    """

    def __init__(
        self,
        dists: Tuple[Distribution, ...],
        eta_levels: VariabilityHierarchy,
        epsilon_levels: VariabilityHierarchy,
    ):
        self._dists = dists
        self._eta_levels = eta_levels
        self._epsilon_levels = epsilon_levels

    @classmethod
    def create(
        cls,
        dists: Optional[Union[Sequence[Distribution], Distribution]] = None,
        eta_levels: Optional[VariabilityHierarchy] = None,
        epsilon_levels: Optional[VariabilityHierarchy] = None,
    ):
        if dists is None:
            dists = ()
        elif isinstance(dists, Distribution):
            dists = (dists,)
        else:
            dists = tuple(dists)
            names = set()
            for dist in dists:
                if not isinstance(dist, Distribution):
                    raise TypeError(f'Can not add variable of type {type(dist)} to RandomVariables')
                for name in dist.names:
                    if name in names:
                        raise ValueError(
                            f'Names of random variables must be unique. Random Variable "{name}" '
                            'was added more than once to RandomVariables'
                        )
                    names.add(name)

        if eta_levels is None:
            iiv_level = VariabilityLevel('IIV', reference=True, group='ID')
            iov_level = VariabilityLevel('IOV', reference=False, group='OCC')
            eta_levels = VariabilityHierarchy((iiv_level, iov_level))
        else:
            if not isinstance(eta_levels, VariabilityHierarchy):
                raise TypeError(
                    f'Type of eta_levels must be a VariabilityHierarchy not a {type(eta_levels)}'
                )

        if epsilon_levels is None:
            ruv_level = VariabilityLevel('RUV', reference=True)
            epsilon_levels = VariabilityHierarchy((ruv_level,))
        else:
            if not isinstance(epsilon_levels, VariabilityHierarchy):
                raise TypeError(
                    f'Type of epsilon_levels must be a VariabilityHierarchy not a {type(epsilon_levels)}'
                )

        return cls(dists, eta_levels, epsilon_levels)

    def replace(self, **kwargs) -> RandomVariables:
        dists = kwargs.get('dists', self._dists)
        eta_levels = kwargs.get('eta_levels', self._eta_levels)
        epsilon_levels = kwargs.get('epsilon_levels', self._epsilon_levels)
        return RandomVariables.create(dists, eta_levels, epsilon_levels)

    @property
    def eta_levels(self) -> VariabilityHierarchy:
        """VariabilityHierarchy for all etas"""
        return self._eta_levels

    @property
    def epsilon_levels(self) -> VariabilityHierarchy:
        """VariabilityHierarchy for all epsilons"""
        return self._epsilon_levels

    def __add__(
        self, other: Union[Distribution, RandomVariables, Sequence[Distribution]]
    ) -> RandomVariables:
        if isinstance(other, Distribution):
            if other.level not in self._eta_levels and other.level not in self._epsilon_levels:
                raise ValueError(
                    "Level of added distribution is not available in any variability hierarchy"
                )
            return RandomVariables(self._dists + (other,), self._eta_levels, self._epsilon_levels)
        elif isinstance(other, RandomVariables):
            if (
                self._eta_levels != other._eta_levels
                or self._epsilon_levels != other._epsilon_levels
            ):
                raise ValueError("RandomVariables must have same variability hierarchies")
            return RandomVariables(
                self._dists + other._dists, self._eta_levels, self._epsilon_levels
            )
        else:
            try:
                dists = tuple(other)
            except TypeError:
                raise TypeError(f'Type {type(other)} cannot be added to RandomVariables')
            else:
                return RandomVariables(self._dists + dists, self._eta_levels, self._epsilon_levels)

    def __radd__(self, other: Union[Distribution, Sequence[Distribution]]) -> RandomVariables:
        if isinstance(other, Distribution):
            if other.level not in self._eta_levels and other.level not in self._epsilon_levels:
                raise ValueError(
                    f"Level {other.level} of added distribution is not available in any variability hierarchy"
                )
            return RandomVariables((other,) + self._dists, self._eta_levels, self._epsilon_levels)
        else:
            try:
                dists = tuple(other)
            except TypeError:
                raise TypeError(f'Type {type(other)} cannot be added to RandomVariables')
            else:
                return RandomVariables(dists + self._dists, self._eta_levels, self._epsilon_levels)

    def __len__(self):
        return len(self._dists)

    @property
    def nrvs(self) -> int:
        n = 0
        for dist in self._dists:
            n += len(dist)
        return n

    def __eq__(self, other):
        if len(self) == len(other):
            for s, o in zip(self._dists, other._dists):
                if s != o:
                    return False
            return (
                self._eta_levels == other._eta_levels
                and self._epsilon_levels == other._epsilon_levels
            )
        return False

    def __hash__(self):
        return hash((self._dists, self._eta_levels, self._epsilon_levels))

    def to_dict(self) -> Dict[str, Any]:
        dists = tuple(d.to_dict() for d in self._dists)
        return {
            'dists': dists,
            'eta_levels': self._eta_levels.to_dict(),
            'epsilon_levels': self._epsilon_levels.to_dict(),
        }

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> RandomVariables:
        eta_levels = VariabilityHierarchy.from_dict(d['eta_levels'])
        epsilon_levels = VariabilityHierarchy.from_dict(d['epsilon_levels'])
        dists = []
        for dist_dict in d['dists']:
            if dist_dict['class'] == 'NormalDistribution':
                dist = NormalDistribution.from_dict(dist_dict)
            else:
                dist = JointNormalDistribution.from_dict(dist_dict)
            dists.append(dist)
        return cls(dists=tuple(dists), eta_levels=eta_levels, epsilon_levels=epsilon_levels)

    def _lookup_rv(self, ind: Union[sympy.Expr, str]):
        if isinstance(ind, sympy.Symbol):
            ind = ind.name
        if isinstance(ind, str):
            for i, dist in enumerate(self._dists):
                if ind in dist.names:
                    return i, dist
        raise KeyError(f'Could not find {ind} in RandomVariables')

    @overload
    def __getitem__(self, ind: Union[int, str, sympy.Expr]) -> Distribution:
        ...

    @overload
    def __getitem__(
        self, ind: Union[slice, Container[Union[str, sympy.Symbol]]]
    ) -> RandomVariables:
        ...

    def __getitem__(self, ind):
        if isinstance(ind, int):
            return self._dists[ind]
        elif isinstance(ind, slice):
            return RandomVariables(
                self._dists[ind.start : ind.stop : ind.step], self._eta_levels, self._epsilon_levels
            )
        elif not isinstance(ind, str) and isinstance(ind, CollectionsContainer):
            remove = [
                name for name in self.names if not ((name in ind) or (sympy.Symbol(name) in ind))
            ]
            split = self.unjoin(remove)
            keep = tuple(
                dist
                for dist in split._dists
                if dist.names[0] in ind or sympy.Symbol(dist.names[0]) in ind
            )
            return RandomVariables(keep, self._eta_levels, self._epsilon_levels)
        else:
            _, rv = self._lookup_rv(ind)
            return rv

    def __contains__(self, ind):
        try:
            self._lookup_rv(ind)
            return True
        except KeyError:
            return False

    @property
    def names(self) -> list[str]:
        """List of the names of all random variables"""
        return list(chain.from_iterable(dist.names for dist in self._dists))

    @property
    def symbols(self) -> list[sympy.Symbol]:
        """List with symbols for all random variables"""
        return [sympy.Symbol(name) for name in self.names]

    @property
    def epsilons(self) -> RandomVariables:
        """Get only the epsilons"""
        return RandomVariables(
            tuple(dist for dist in self._dists if dist.level in self._epsilon_levels.names),
            self._eta_levels,
            self._epsilon_levels,
        )

    @property
    def etas(self) -> RandomVariables:
        """Get only the etas"""
        return RandomVariables(
            tuple(dist for dist in self._dists if dist.level in self._eta_levels.names),
            self._eta_levels,
            self._epsilon_levels,
        )

    @property
    def iiv(self) -> RandomVariables:
        """Get only the iiv etas, i.e. etas with variability level 0"""
        return RandomVariables(
            tuple(dist for dist in self._dists if dist.level == self._eta_levels[0].name),
            self._eta_levels,
            self._epsilon_levels,
        )

    @property
    def iov(self) -> RandomVariables:
        """Get only the iov etas, i.e. etas with variability level 1"""
        return RandomVariables(
            tuple(dist for dist in self._dists if dist.level == self._eta_levels[1].name),
            self._eta_levels,
            self._epsilon_levels,
        )

    @property
    def free_symbols(self) -> Set[sympy.Expr]:
        """Set of free symbols for all random variables"""
        return set().union(*(dist.free_symbols for dist in self._dists))

    @property
    def parameter_names(self) -> Tuple[str, ...]:
        """List of parameter names for all random variables"""
        params = set().union(*(dist.parameter_names for dist in self._dists))
        return tuple(sorted(map(str, params)))

    @property
    def variance_parameters(self) -> List[str]:
        """List of all parameters representing variance for all random variables"""
        parameters = []
        for dist in self._dists:
            if isinstance(dist, NormalDistribution):
                p = dist.variance
                if p not in parameters:
                    parameters.append(p)
            else:
                assert isinstance(dist, JointNormalDistribution)
                for p in dist.variance.diagonal():
                    if p not in parameters:
                        parameters.append(p)
        return [p.name for p in parameters]

    def get_covariance(
        self, rv1: Union[sympy.Symbol, str], rv2: Union[sympy.Symbol, str]
    ) -> sympy.Expr:
        """Get covariance between two random variables"""
        _, dist1 = self._lookup_rv(rv1)
        _, dist2 = self._lookup_rv(rv2)
        if dist1 is not dist2:
            return sympy.Integer(0)
        else:
            name1 = rv1.name if not isinstance(rv1, str) else rv1
            name2 = rv2.name if not isinstance(rv2, str) else rv2
            return dist1.get_covariance(name1, name2)

    def subs(self, d: Mapping[Union[str, sympy.Expr], sympy.Expr]) -> RandomVariables:
        """Substitute expressions

        Parameters
        ----------
        d : dict
            Dictionary of from: to pairs for substitution

        Examples
        --------
        >>> import sympy
        >>> from pharmpy.model import RandomVariables, Parameter
        >>> omega = Parameter("OMEGA_CL", 0.1)
        >>> dist = NormalDistribution.create("IIV_CL", "IIV", 0, omega.symbol)
        >>> rvs = RandomVariables.create([dist])
        >>> rvs.subs({omega.symbol: sympy.Symbol("OMEGA_NEW")})
        IIV_CL ~ N(0, OMEGA_NEW)

        """
        new_dists = tuple(dist.subs(d) for dist in self._dists)
        return self.replace(dists=new_dists)

    def unjoin(
        self, inds: Union[str, sympy.Symbol, Iterable[Union[str, sympy.Symbol]]]
    ) -> RandomVariables:
        """Remove all covariances the random variables have with other random variables

        Parameters
        ----------
        inds
            One or multiple names or symbols to unjoin

        Examples
        --------
        >>> from pharmpy.model import RandomVariables, JointNormalDistribution, Parameter
        >>> omega_cl = Parameter("OMEGA_CL", 0.1)
        >>> omega_v = Parameter("OMEGA_V", 0.1)
        >>> corr_cl_v = Parameter("OMEGA_CL_V", 0.01)
        >>> dist1 = JointNormalDistribution.create(["IIV_CL", "IIV_V"], 'IIV', [0, 0],
        ...     [[omega_cl.symbol, corr_cl_v.symbol], [corr_cl_v.symbol, omega_v.symbol]])
        >>> rvs = RandomVariables.create([dist1])
        >>> rvs.unjoin('IIV_CL')
        IIV_CL ~ N(0, OMEGA_CL)
        IIV_V ~ N(0, OMEGA_V)

        See Also
        --------
        join
        """
        if isinstance(inds, str) or isinstance(inds, sympy.Symbol):
            inds = [inds]
        inds = [ind.name if not isinstance(ind, str) else ind for ind in inds]

        newdists = []
        for dist in self._dists:
            first = True
            keep = None
            if isinstance(dist, JointNormalDistribution) and any(
                item in dist.names for item in inds
            ):
                for i, name in enumerate(dist.names):
                    if name in inds:  # unjoin  this
                        mean = dist.mean[i]
                        assert isinstance(mean, sympy.Expr)
                        variance = dist.variance[i, i]
                        assert isinstance(variance, sympy.Expr)
                        new = NormalDistribution(name, dist.level, mean, variance)
                        newdists.append(new)
                    elif first:  # first of the ones to keep
                        first = False
                        remove = [i for i, n in enumerate(dist.names) if n in inds]
                        if len(dist) - len(remove) == 1:
                            mean = dist.mean[i]
                            assert isinstance(mean, sympy.Expr)
                            variance = dist.variance[i, i]
                            assert isinstance(variance, sympy.Expr)
                            keep = NormalDistribution(name, dist.level, mean, variance)
                        else:
                            names = list(dist.names)
                            mean = sympy.Matrix(dist.mean)
                            variance = sympy.Matrix(dist.variance)
                            for i in reversed(remove):
                                del names[i]
                                mean.row_del(i)
                                variance.row_del(i)
                                variance.col_del(i)
                            keep = JointNormalDistribution(
                                tuple(names),
                                dist.level,
                                sympy.ImmutableMatrix(mean),
                                sympy.ImmutableMatrix(variance),
                            )
                if keep is not None:
                    newdists.append(keep)
            else:
                newdists.append(dist)
        new_rvs = RandomVariables(tuple(newdists), self._eta_levels, self._epsilon_levels)
        return new_rvs

    def join(
        self,
        inds: Collection[Union[str, sympy.Symbol]],
        fill: Union[int, float, sympy.Expr] = 0,
        name_template: Optional[str] = None,
        param_names: Optional[list[str]] = None,
    ) -> Tuple[RandomVariables, Dict[str, Tuple[str, str]]]:
        """Join random variables together into one joint distribution

        Set new covariances (and previous 0 covs) to 'fill'.
        All joined random variables will form a new joint normal distribution and
        if they were part of previous joint normal distributions they will be taken out
        from these and the remaining variables will be stay.

        Parameters
        ----------
        inds
            Indices of variables to join
        fill : value
            Value to use for new covariances. Default is 0
        name_template : str
            A string template to use for new covariance symbols.
            Using this option will override fill.
        param_names : list
            List of parameter names to be used together with
            name_template.

        Returns
        -------
        A tuple of a the new RandomVariables and a dictionary from newly created covariance parameter names to
        tuple of parameter names. Empty dictionary if no parameter symbols were created

        Examples
        --------
        >>> from pharmpy.model import RandomVariables, NormalDistribution, Parameter
        >>> omega_cl = Parameter("OMEGA_CL", 0.1)
        >>> omega_v = Parameter("OMEGA_V", 0.1)
        >>> dist1 = NormalDistribution.create("IIV_CL", "IIV", 0, omega_cl.symbol)
        >>> dist2 = NormalDistribution.create("IIV_V", "IIV", 0, omega_v.symbol)
        >>> rvs = RandomVariables.create([dist1, dist2])
        >>> rvs, _ = rvs.join(['IIV_CL', 'IIV_V'])
        >>> rvs
        ⎡IIV_CL⎤    ⎧⎡0⎤  ⎡OMEGA_CL     0   ⎤⎫
        ⎢      ⎥ ~ N⎪⎢ ⎥, ⎢                 ⎥⎪
        ⎣IIV_V ⎦    ⎩⎣0⎦  ⎣   0      OMEGA_V⎦⎭

        See Also
        --------
        unjoin
        """
        if any(item not in self.names for item in inds):
            raise KeyError("Cannot join non-existing random variable")
        joined_rvs = self[inds]
        assert isinstance(joined_rvs, RandomVariables)
        means, M, names = joined_rvs._calc_covariance_matrix()
        cov_to_params = {}
        if fill != 0:
            for row, col in product(range(M.rows), range(M.cols)):
                if M[row, col] == 0:
                    M[row, col] = fill
        elif name_template:
            for row, col in product(range(M.rows), range(M.cols)):
                if M[row, col] == 0 and row > col:
                    param_1, param_2 = M[row, row], M[col, col]
                    assert isinstance(param_names, list)
                    cov_name = name_template.format(param_names[col], param_names[row])
                    cov_to_params[cov_name] = (str(param_1), str(param_2))
                    M[row, col], M[col, row] = sympy.Symbol(cov_name), sympy.Symbol(cov_name)
        joined_dist = JointNormalDistribution(
            tuple(names),
            joined_rvs[0].level,
            sympy.ImmutableMatrix(means),
            sympy.ImmutableMatrix(M),
        )

        unjoined_rvs = self.unjoin(inds)
        newdists = []
        first = True
        for dist in unjoined_rvs._dists:
            if any(item in dist.names for item in inds):
                if first:
                    first = False
                    newdists.append(joined_dist)
            else:
                newdists.append(dist)

        new_rvs = RandomVariables(tuple(newdists), self._eta_levels, self._epsilon_levels)
        return new_rvs, cov_to_params

    def nearest_valid_parameters(self, parameter_values: Mapping[str, float]) -> dict[str, float]:
        """Force parameter values into being valid

        As small changes as possible

        returns a dict with the valid parameter values
        """
        nearest = dict(parameter_values)  # copy
        for dist in self._dists:
            if isinstance(dist, JointNormalDistribution):
                symb_sigma = dist.variance
                sigma = symb_sigma.subs(dict(parameter_values))
                A = np.array(sigma).astype(np.float64)
                B = nearest_positive_semidefinite(A)
                if B is not A:
                    for row in range(len(A)):
                        for col in range(row + 1):
                            elt = symb_sigma[row, col]
                            assert isinstance(elt, sympy.Symbol)
                            nearest[elt.name] = B[row, col]
        return nearest

    def validate_parameters(self, parameter_values: Mapping[str, float]) -> bool:
        """Validate a dict or Series of parameter values

        Currently checks that all covariance matrices are positive semidefinite
        use_cache for using symengine cached matrices
        """
        for dist in self._dists:
            if isinstance(dist, JointNormalDistribution):
                sigma = dist._symengine_variance
                replacement = {}
                for param in dict(parameter_values):
                    replacement[symengine.Symbol(param)] = parameter_values[param]
                sigma = sigma.subs(replacement)
                if not sigma.free_symbols:  # Cannot validate since missing params
                    a = np.array(sigma).astype(np.float64)
                    if not is_positive_semidefinite(a):
                        return False
                else:
                    raise TypeError("Cannot validate parameters since all are not numeric")
        return True

    def sample(
        self,
        expr,
        parameters: Optional[Mapping[str, float]] = None,
        samples: int = 1,
        rng: Optional[np.random.Generator] = None,
    ) -> np.ndarray:
        """Sample from the distribution of expr

        parameters in the distributions will first be replaced"""

        sympified_expr = parse_expr(expr)
        xreplace_parameters = {} if parameters is None else xreplace_dict(parameters)

        return _sample_from_distributions(
            self,
            sympified_expr,
            xreplace_parameters,
            samples,
            _create_rng(rng),
        )

    def _calc_covariance_matrix(self) -> Tuple[List[sympy.Expr], sympy.ImmutableMatrix, List[str]]:
        means = []
        names = []
        n = 0
        for dist in self._dists:
            names.extend(dist.names)
            n += len(dist)

        M = sympy.zeros(n)

        if not names:
            return means, M, names

        row = 0
        col = 0
        for dist in self._dists:
            if isinstance(dist, NormalDistribution):
                means.append(dist.mean)
                M[row, col] = dist.variance
                row += 1
                col += 1
            else:
                assert isinstance(dist, JointNormalDistribution)
                means.extend(dist.mean)  # pyright: ignore [reportGeneralTypeIssues]
                var = dist.variance
                for i in range(var.rows):
                    for j in range(var.cols):
                        M[row + i, col + j] = var[i, j]
                row += var.rows
                col += var.cols
        return means, M, names

    @property
    def covariance_matrix(self) -> sympy.ImmutableMatrix:
        """Covariance matrix of all random variables"""
        _, M, _ = self._calc_covariance_matrix()
        return M

    def __repr__(self):
        return '\n'.join(map(repr, self._dists))

    def _repr_latex_(self):
        lines = (dist.latex_string(aligned=True) for dist in self._dists)
        return '\\begin{align*}\n' + r' \\ '.join(lines) + '\\end{align*}'

    def parameters_sdcorr(self, values: Mapping[str, float]) -> dict[str, float]:
        """Convert parameter values to sd/corr form

        All parameter values will be converted to sd/corr assuming
        they are given in var/cov form.

        Parameters
        ----------
        values : dict
            Dict of parameter names to values
        """
        newdict = dict(values)
        for dist in self._dists:
            if isinstance(dist, JointNormalDistribution):
                sigma_sym = dist.variance
                sigma = np.array(sigma_sym.subs(values)).astype(np.float64)
                corr = cov2corr(sigma)
                for i in range(sigma_sym.rows):
                    for j in range(sigma_sym.cols):
                        elt = sigma_sym[i, j]
                        assert isinstance(elt, sympy.Symbol)
                        name = elt.name
                        if i != j:
                            newdict[name] = corr[i, j]
                        else:
                            newdict[name] = np.sqrt(sigma[i, j])
            else:
                assert isinstance(dist, NormalDistribution)
                variance = dist.variance
                if isinstance(variance, sympy.Symbol):
                    name = variance.name
                else:
                    raise NotImplementedError(
                        "parameters_sdcorr only supports pure" " symbols as variances"
                    )
                if name in newdict:
                    newdict[name] = float(
                        np.sqrt(np.array(subs(variance, values)).astype(np.float64))
                    )
        return newdict

    def get_rvs_with_same_dist(self, rv: Union[str, sympy.Symbol]) -> RandomVariables:
        """Gets random variables with the same distribution as input random variable

        The resulting RandomVariables objects includes the input random variable.

        Parameters
        ----------
        rv : str
            Name of random variable

        Returns
        -------
        RandomVariables
            RandomVariables object with all distributions as input random variable (including input)
        """
        _, dist_input = self._lookup_rv(rv)

        rvs = [dist for dist in self if dist.variance == dist_input.variance]

        return RandomVariables.create(rvs)

    def replace_with_sympy_rvs(self, expr: sympy.Expr) -> sympy.Expr:
        """Replaces Pharmpy RVs in a Sympy expression with Sympy RVs

        Takes a Sympy expression and replaces all RVs with Sympy RVs, resulting expression
        can be used in different Sympy functions (e.g. sympy.stats.std())

        Parameters
        ----------
        expr : sympy.Expr
            Expression which will get RVs replaced

        Returns
        -------
        sympy.Expr
            Expression with replaced RVs
        """
        rvs_in_expr = {
            self[rv]  # pyright: ignore [reportGeneralTypeIssues]
            for rv in expr.free_symbols.intersection(self.symbols)
        }
        subs_dict = {}
        for i, rv in enumerate(rvs_in_expr):
            if isinstance(rv, JointNormalDistribution):
                # sympy.stats.MultivariateNormal uses variance, sympy.stats.Normal takes std
                dist = sympy_stats.MultivariateNormal(f'__rv{i}', rv.mean, rv.variance)
                for j in range(0, len(rv.names)):
                    subs_dict[rv.names[j]] = dist[j]  # pyright: ignore [reportGeneralTypeIssues]
            else:
                subs_dict[rv.names[0]] = sympy_stats.Normal(
                    f'__rv{i}', rv.mean, sympy.sqrt(rv.variance)
                )
        return expr.subs(subs_dict)  # pyright: ignore [reportGeneralTypeIssues]


def _sample_from_distributions(distributions, expr, parameters, nsamples, rng):
    random_variable_symbols = expr.free_symbols.difference(parameters.keys())
    filtered_distributions = filter_distributions(distributions, random_variable_symbols)
    sampling_rvs = subs_distributions(filtered_distributions, parameters)
    sampled_expr = subs(expr, parameters, simultaneous=True)
    return sample_expr_from_rvs(sampling_rvs, sampled_expr, nsamples, rng)


def filter_distributions(
    distributions: Iterable[Distribution], symbols: Set[sympy.Symbol]
) -> Iterable[Distribution]:
    covered_symbols = set()

    for dist in distributions:
        rvs_covered_by_dist = tuple(rv for rv in dist.names if sympy.Symbol(rv) in symbols)
        if rvs_covered_by_dist:
            yield dist[rvs_covered_by_dist]
            covered_symbols.update(sympy.Symbol(rv) for rv in rvs_covered_by_dist)

    if covered_symbols != symbols:
        raise ValueError('Could not cover all requested symbols with given distributions')


def subs_distributions(
    distributions: Iterable[Distribution], parameters: Dict[sympy.Symbol, float]
) -> Iterable[Tuple[Tuple[sympy.Symbol, ...], NumericDistribution]]:
    for dist in distributions:
        rvs_symbols = tuple(map(sympy.Symbol, dist.names))
        numeric_distribution = dist.evalf(parameters)
        yield (rvs_symbols, numeric_distribution)


def sample_expr_from_rvs(
    sampling_rvs: Iterable[Tuple[Tuple[sympy.Expr, ...], NumericDistribution]],
    expr: sympy.Expr,
    nsamples: int,
    rng,
):
    samples = sample_rvs(sampling_rvs, nsamples, rng)
    return eval_expr(expr, nsamples, samples)


def sample_rvs(
    sampling_rvs: Iterable[Tuple[Tuple[sympy.Expr, ...], NumericDistribution]],
    nsamples: int,
    rng,
) -> Dict[sympy.Expr, np.ndarray]:
    data = {}
    for symbols, distribution in sampling_rvs:
        cursample = distribution.sample(rng, nsamples)
        if len(symbols) > 1:
            # NOTE: This makes column iteration faster
            cursample = np.array(cursample, order='F')
            for j, s in enumerate(symbols):
                data[s] = cursample[:, j]
        else:
            data[symbols[0]] = cursample

    return data



import re
from typing import Set, Union

import pharmpy.model
from pharmpy.deps import sympy, sympy_printing
from pharmpy.internals.code_generator import CodeGenerator
from pharmpy.model import Assignment, Infusion
from pharmpy.modeling import get_bioavailability, get_lag_times

from .error_model import res_error_term
from .name_mangle import name_mangle


class ExpressionPrinter(sympy_printing.str.StrPrinter):
    def __init__(self, amounts):
        self.amounts = amounts
        super().__init__()

    def _print_Symbol(self, expr):
        return name_mangle(expr.name)

    def _print_Derivative(self, expr):
        fn = expr.args[0]
        return f'd/dt({fn.name})'

    def _print_Function(self, expr):
        name = expr.func.__name__
        if name in self.amounts:
            return expr.func.__name__
        else:
            return expr.func.__name__ + f'({self.stringify(expr.args, ", ")})'


def add_statements(
    model: pharmpy.model.Model,
    cg: CodeGenerator,
    statements: pharmpy.model.statements,
    only_piecewise: Union[bool, None] = None,
    dependencies: Set[sympy.Symbol] = set(),
    res_alias: Set[sympy.Symbol] = set(),
):
    """
    Will add the provided statements to the code generator objects, translated
    to nlmixr format.

    Parameters
    ----------
    model : pharmpy.model.Model
        A pharmpy model object to add statements to.
    cg : CodeGenerator
        Codegenerator object holding the code to be added to.
    statements : pharmpy.model.statements
        Statements to be added to the code generator.
    only_piecewise : Union[bool, None], optional
        Is the dependent variable only dependent on piecewise statements.
        The default is None.
    dependencies : set[sympy.Symbols], optional
        A set with symbols that the dependent variable are dependent on.
        Could for instance a term 'W' which define the error model.
        The default is set().
    res_alias : set[sympy.Symbols], optional
        A set with aliases for the dependent or resulting term for the
        dependent variable. The default is set().

    """

    # FIXME: Handle other DVs?
    dv = list(model.dependent_variables.keys())[0]

    for s in statements:
        if isinstance(s, Assignment):
            if s.symbol == dv and not s.expression.is_Piecewise:
                # FIXME: Find another way to assert that a sigma exist
                sigma = None
                for dist in model.random_variables.epsilons:
                    sigma = dist.variance
                assert sigma is not None

                if only_piecewise is False:
                    dv_term = res_error_term(model, s.expression)
                    res = dv_term.res
                    if len(dependencies) != 0:
                        cg.add(f'{s.symbol} <- {res}')
                        cg.add(f'{s.symbol} ~ add(add_error) + prop(prop_error)')

                        # TODO: Remove sigma here instead of in ini
                        # also remove aliases

                    else:
                        cg.add(f"{s.symbol} <- {res}")
                        cg.add(f'add_error <- {dv_term.add.expr}')
                        cg.add(f'prop_error <- {dv_term.prop.expr}')
                        cg.add(f'{s.symbol} ~ add(add_error) + prop(prop_error)')
                else:
                    dv_term = res_error_term(model, s.expression)
                    cg.add(f"res <- {dv_term.res}")
                    cg.add(f'add_error <- {dv_term.add.expr}')
                    cg.add(f'prop_error <- {dv_term.prop.expr}')

            elif model.statements.ode_system is not None and (
                s.symbol in get_bioavailability(model).values()
                or s.symbol in get_lag_times(model).values()
            ):
                pass
            else:
                expr = s.expression
                if expr.is_Piecewise:
                    first = True
                    for value, cond in expr.args:
                        if cond is not sympy.S.true:
                            cond = convert_eq(cond)
                            if first:
                                cg.add(f'if ({cond}) {{')
                                first = False
                            else:
                                cg.add(f'}} else if ({cond}) {{')
                        else:
                            cg.add('} else {')
                            if "NEWIND" in [t.name for t in expr.free_symbols] and value == 0:
                                largest_value = expr.args[0].expr
                                largest_cond = expr.args[0].cond
                                for value, cond in expr.args[1:]:
                                    if cond is not sympy.S.true:
                                        if cond.rhs > largest_cond.rhs:
                                            largest_value = value
                                            largest_cond = cond
                                        elif cond.rhs == largest_cond.rhs:
                                            if not isinstance(cond, sympy.LessThan) and isinstance(
                                                largest_cond, sympy.LessThan
                                            ):
                                                largest_value = value
                                                largest_cond = cond
                                value = largest_value
                        cg.indent()

                        if only_piecewise is False:
                            cg.add(f'{s.symbol.name} <- {value}')

                            if s.symbol in dependencies:
                                if not value.is_constant() and not isinstance(value, sympy.Symbol):
                                    add, prop = extract_add_prop(value, res_alias, model)
                                    cg.add(f'add_error <- {add}')
                                    cg.add(f'prop_error <- {prop}')
                        elif s.symbol == dv:
                            if value != dv:
                                t = res_error_term(model, value)
                                # FIXME: Remove sigma here instead of in ini
                                # also remove aliases for sigma
                                cg.add(f"res <- {t.res}")
                                cg.add(f'add_error <- {t.add.expr}')
                                cg.add(f'prop_error <- {t.prop.expr}')
                            else:
                                cg.add("res <- res")
                                cg.add('add_error <- add_error')
                                cg.add('prop_error <- prop_error')
                        else:
                            cg.add(f'{s.symbol.name} <- {value}')

                        cg.dedent()
                    cg.add('}')
                elif s.symbol in dependencies:
                    add, prop = extract_add_prop(s.expression, res_alias, model)
                    cg.add(f'{s.symbol.name} <- {expr}')  # TODO: Remove?
                    cg.add(f'add_error <- {add}')
                    cg.add(f'prop_error <- {prop}')
                else:
                    cg.add(f'{s.symbol.name} <- {expr}')

    if only_piecewise:
        cg.add(f'{dv} <- res')
        cg.add(f'{dv} ~ add(add_error) + prop(prop_error)')


def extract_add_prop(s, res_alias: Set[sympy.symbols], model: pharmpy.model.Model):
    """
        Extract additive and proportional error terms from a sympy expression

        Parameters
        ----------
        s : A sympy expression
            A sympy expression from which an additive and a proportional error
            term are to be extracted.
        res_alias : set[sympy.symbols]
            A set with aliases for the dependent or resulting term for the
            dependent variable.
        model : pharmpy.model.Model
            The connected pharmpy model.

    Returns
        -------
        add : sympy.Symbol
            The symbol representing the additive error. Zero if none found
        prop : sympy.Symbol
            The symbol representing  the proportional error. Zero if none found

    """
    if isinstance(s, sympy.Symbol):
        terms = [s]
    elif isinstance(s, sympy.Pow):
        terms = sympy.Add.make_args(s.args[0])
    elif isinstance(s, sympy.Mul):
        terms = [s]
    elif isinstance(s, sympy.Integer):
        terms = [s]
    elif isinstance(s, sympy.Float):
        terms = [s]
    else:
        terms = sympy.Add.make_args(s.expression)
    assert len(terms) <= 2

    w = False

    if isinstance(s, sympy.Pow):
        s_arg = sympy.Add.make_args(s.args[0])
        if len(s_arg) <= 2:
            all_pow = True
            for t in s_arg:
                for f in sympy.Mul.make_args(t):
                    if (
                        isinstance(f, sympy.Pow)
                        or isinstance(f, sympy.Integer)
                        or isinstance(f, sympy.Float)
                    ):
                        pass
                    else:
                        all_pow = False
        if all_pow:
            w = True

    prop = 0
    add = 0
    prop_found = False
    for term in terms:
        for symbol in term.free_symbols:
            if symbol in res_alias:
                if prop_found is False:
                    term = term.subs(symbol, 1)
                    if w:
                        prop = sympy.sqrt(term)
                    else:
                        prop += term
                    prop_found = True
        if prop_found is False:
            if w:
                add = sympy.sqrt(term)
            else:
                add += term
    return add, prop


def add_bio_lag(model: pharmpy.model.Model, cg: CodeGenerator, bio=False, lag=False):
    if bio:
        bio_lag = get_bioavailability(model)
    elif lag:
        bio_lag = get_lag_times(model)
    else:
        return

    for s in model.statements.before_odes:
        if s.symbol in bio_lag.values():
            comp = list(bio_lag.keys())[list(bio_lag.values()).index(s.symbol)]

            if s.expression.is_Piecewise:
                first = True
                for value, cond in s.expression.args:
                    if cond is not sympy.S.true:
                        cond = convert_eq(cond)
                        if first:
                            cg.add(f'if ({cond}) {{')
                            first = False
                        else:
                            cg.add(f'}} else if ({cond}) {{')
                    else:
                        cg.add('} else {')

                    if bio:
                        cg.add(f'f(A_{comp}) <- {value}')
                    elif lag:
                        cg.add(f'alag(A_{comp}) <- {value}')
            else:
                if bio:
                    cg.add(f'f(A_{comp}) <- {s.expression}')
                elif lag:
                    cg.add(f'alag(A_{comp}) <- {s.expression}')


def add_piecewise(model: pharmpy.model.Model, cg: CodeGenerator, s):
    expr = s.expression
    first = True
    for value, cond in expr.args:
        if cond is not sympy.S.true:
            if cond.atoms(sympy.Eq):
                cond = convert_eq(cond)
            if first:
                cg.add(f'if ({cond}) {{')
                first = False
            else:
                cg.add(f'}} else if ({cond}) {{')
        else:
            cg.add('} else {')
            if "NEWIND" in [t.name for t in expr.free_symbols] and value == 0:
                largest_value = expr.args[0].expr
                largest_cond = expr.args[0].cond
                for value, cond in expr.args[1:]:
                    if cond is not sympy.S.true:
                        if cond.rhs > largest_cond.rhs:
                            largest_value = value
                            largest_cond = cond
                        elif cond.rhs == largest_cond.rhs:
                            if not isinstance(cond, sympy.LessThan) and isinstance(
                                largest_cond, sympy.LessThan
                            ):
                                largest_value = value
                                largest_cond = cond
                value = largest_value
        cg.indent()
        cg.add(f'{s.symbol.name} <- {value}')
        cg.dedent()
    cg.add('}')


def add_ode(model: pharmpy.model.Model, cg: CodeGenerator) -> None:
    """
    Add the ODEs from a model to a code generator

    Parameters
    ----------
    model : pharmpy.model.Model
        A pharmpy model object to add ODEs to.
    cg : CodeGenerator
        Codegenerator object holding the code to be added to.
    """
    amounts = [am.name for am in list(model.statements.ode_system.amounts)]
    printer = ExpressionPrinter(amounts)

    for eq in model.statements.ode_system.eqs:
        # Should remove piecewise from these equations in nlmixr
        if eq.atoms(sympy.Piecewise):
            lhs = remove_piecewise(printer.doprint(eq.lhs))
            rhs = remove_piecewise(printer.doprint(eq.rhs))

            cg.add(f'{lhs} = {rhs}')
        else:
            cg.add(f'{printer.doprint(eq.lhs)} = {printer.doprint(eq.rhs)}')

    for comp in model.statements.ode_system.dosing_compartments:
        # FIXME : Handle multiple doses with different dur/rate
        dose = comp.doses[0]
        if isinstance(dose, Infusion):
            if dur := dose.duration:
                cg.add(f'dur({comp.amount.name}) = {dur}')
            if rate := dose.rate:
                cg.add(f'rate({comp.amount.name}) = {rate}')


def remove_piecewise(expr: str) -> str:
    """
    Return an expression without Piecewise statements

    Parameters
    ----------
    expr : str
        A sympy expression, given as a string

    Returns
    -------
    str
        Return given expression but without Piecewise statements

    """
    all_piecewise = find_piecewise(expr)
    # Go into each piecewise found
    for p in all_piecewise:
        expr = piecewise_replace(expr, p, "")
    return expr


def find_piecewise(expr: str) -> list:
    """
    Locate all Piecewise statements in an expression and return them as a list

    Parameters
    ----------
    expr : str
        A sympy expression in string form

    Returns
    -------
    list
        A list of all piecewise statements found in the expression

    """
    d = find_parentheses(expr)

    piecewise_start = [m.start() + len("Piecewise") for m in re.finditer("Piecewise", expr)]

    all_piecewise = []
    for p in piecewise_start:
        if p in d:
            all_piecewise.append(expr[p + 1 : d[p]])
    return all_piecewise


def find_parentheses(s: str) -> dict:
    """
    Find all matching parenthesis in a given string

    Parameters
    ----------
    s : str
        Any string

    Returns
    -------
    dict
        A dictionary with keys corresponding to positions of the opening
        parenthesis and value being the position of the closing parenthesis.

    """
    start = []  # all opening parentheses
    d = {}

    for i, c in enumerate(s):
        if c == '(':
            start.append(i)
        if c == ')':
            try:
                d[start.pop()] = i
            except IndexError:
                print('Too many closing parentheses')
    if start:  # Check if stack is empty afterward
        print('Too many opening parentheses')

    return d


def piecewise_replace(expr: str, piecewise: sympy.Piecewise, s: str) -> str:
    """
    Will replace a given piecewise expression with wanted string, s

    Parameters
    ----------
    expr : str
        A string representing a sympy expression
    piecewise : sympy.Piecewise
        A sympy Piecewise statement to be changed
    s : str
        A string with wanted information to change piecewise to

    Returns
    -------
    str
        Expression with replaced piecewise expressions.

    """
    if s == "":
        expr = re.sub(r'([\+\-\/\*]\s*)(Piecewise)', r'\2', expr)
        return expr.replace(f'Piecewise({piecewise})', s)
    else:
        return expr.replace(f'Piecewise({piecewise})', s)


def convert_eq(cond: sympy.Eq) -> str:
    """
    Convert a sympy equal statement to R syntax

    Parameters
    ----------
    cond : sympy.Eq
        Sympy equals statement

    Returns
    -------
    str
        A string with R format for the same statement

    """
    cond = sympy.pretty(cond)
    cond = cond.replace("=", "==")
    cond = cond.replace("≠", "!=")
    cond = cond.replace("≤", "<=")
    cond = cond.replace("≥", "<=")
    cond = cond.replace("∧", "&")
    cond = cond.replace("∨", "|")

    cond = re.sub(r'(ID\s*==\s*)(\d+)', r"\1'\2'", cond)

    if (
        re.search(r'(ID\s*<=\s*)(\d+)', cond)
        or re.search(r'(ID\s*>=\s*)(\d+)', cond)
        or re.search(r'(ID\s*<\s*)(\d+)', cond)
        or re.search(r'(ID\s*>\s*)(\d+)', cond)
    ):
        print(f"Condition '{cond}' not supported by nlmixr. Model will not run.")
    return cond



from typing import Mapping, Tuple, Union

from lark import Tree, Visitor
from lark.lexer import Token


def _rule_of(item: Union[Tree, Token]) -> str:
    if isinstance(item, Tree):
        return item.data
    else:
        return item.type


def _empty_node(rule: str):
    """Create empty Tree or Token, depending on (all) upper-case or not."""
    return Token(rule, '') if rule == rule.upper() else Tree(rule, [])


class InsertMissing(Visitor):
    def __init__(self, missing: Tuple[Mapping[str, Tuple[Tuple[int, str], ...]], ...]):
        self._missing = missing

    def __default__(self, tree: Tree):
        for d in self._missing:
            try:
                children = d[tree.data]
            except KeyError:
                pass
            else:
                for pos, name in children:
                    if not any(_rule_of(child) == name for child in tree.children):
                        tree.children.insert(pos, _empty_node(name))



from functools import partial
from itertools import product
from typing import Dict, Iterable, List, Sequence, Tuple, TypeVar, Union

from pharmpy.deps import sympy
from pharmpy.model import Model
from pharmpy.modeling import get_bioavailability
from pharmpy.modeling.covariate_effect import (
    EffectType,
    OperationType,
    add_covariate_effect,
    remove_covariate_effect,
)
from pharmpy.modeling.expressions import (
    get_individual_parameters,
    get_parameter_rv,
    get_pd_parameters,
    get_pk_parameters,
)

from ..statement.definition import Let
from ..statement.feature.covariate import (
    Covariate,
    CovariateWildcard,
    EffectFunctionWildcard,
    ParameterWildcard,
    Ref,
)
from ..statement.feature.symbols import Symbol, Wildcard
from ..statement.statement import Statement
from .feature import Feature

T = TypeVar('T')

EffectLiteral = Tuple[str, str, EffectType, OperationType]
EffectSpecFeature = Union[T, Tuple[T, ...]]
InputEffectSpecFeature = Union[T, Sequence[T]]
Spec = Tuple[
    EffectSpecFeature[str],
    EffectSpecFeature[str],
    EffectSpecFeature[EffectType],
    EffectSpecFeature[OperationType],
]
InputSpec = Sequence[InputEffectSpecFeature[Union[str, EffectType, OperationType]],]

all_continuous_covariate_effects = (
    'lin',
    'piece_lin',
    'exp',
    'pow',
)

all_categorical_covariate_effects = ('cat',)

all_covariate_effects = all_continuous_covariate_effects + all_categorical_covariate_effects


def features(model: Model, statements: Iterable[Statement], remove=False) -> Iterable[Feature]:
    # Add remove_covariate_effect if optional argument
    for args in parse_spec(spec(model, statements)):
        if remove or args[-1]:
            yield ('COVARIATE',) + args[:-1] + ('REMOVE',), partial(
                remove_covariate_effect, parameter=args[0], covariate=args[1]
            )
        if not remove and (args[-1] or not args[-1]):
            yield ('COVARIATE',) + args[:-1] + ('ADD',), partial(
                add_covariate_effect,
                parameter=args[0],
                covariate=args[1],
                effect=args[2],
                operation=args[3],
            )


Definitions = Dict[str, Tuple[str, ...]]


def _partition_statements(statements: Iterable[Statement]) -> Tuple[List[Covariate], Definitions]:
    effects = []
    definitions = {}
    for statement in statements:
        if isinstance(statement, Covariate):
            effects.append(statement)
        elif isinstance(statement, Let):
            definitions[statement.name] = statement.value

    return effects, definitions


def spec(model: Model, statements: Iterable[Statement]) -> Iterable[Spec]:
    effects, definitions = _partition_statements(statements)

    for effect in effects:
        t = _effect_to_tuple(model, definitions, effect)
        if all([e != tuple() for e in t]):  # NOTE: We do not yield empty products
            yield t


def covariates(model: Model, statements: Iterable[Statement]) -> Iterable[str]:
    # NOTE: This yields the covariates present in the COVARIATE statements
    effects, definition = _partition_statements(statements)

    for effect in effects:
        yield from _effect_to_covariates(model, definition, effect)


def _ensure_tuple_or_list(x):
    return x if isinstance(x, (tuple, list)) else (x,)


def parse_spec(spec: Iterable[Spec]) -> Iterable[EffectLiteral]:
    for parameters, covariates, fps, operations, optional in spec:
        parameters = _ensure_tuple_or_list(parameters)
        covariates = _ensure_tuple_or_list(covariates)
        fps = _ensure_tuple_or_list(fps)
        operations = _ensure_tuple_or_list(operations)
        optional = _ensure_tuple_or_list(optional)
        yield from product(parameters, covariates, fps, operations, optional)


def _effect_to_tuple(model: Model, definitions: Definitions, effect: Covariate) -> Spec:
    parameters = (
        _interpret_symbol(model, definitions, effect.parameter)
        if isinstance(effect.parameter, Symbol)
        else effect.parameter
    )
    covariates = _effect_to_covariates(model, definitions, effect)
    fps = all_continuous_covariate_effects if effect.fp is EffectFunctionWildcard else effect.fp
    ops = effect.op
    optional = effect.optional.option
    return (parameters, covariates, tuple(fp.lower() for fp in fps), ops, optional)


def _effect_to_covariates(
    model: Model, definition: Definitions, effect: Covariate
) -> Tuple[str, ...]:
    return (
        _interpret_symbol(model, definition, effect.covariate)
        if isinstance(effect.covariate, Symbol)
        else effect.covariate
    )


def _interpret_symbol(model: Model, definition, symbol: Symbol) -> Tuple[str, ...]:
    if isinstance(symbol, Ref):
        try:
            return definition[symbol.name]
        except KeyError:
            return _interpret_ref(model, symbol)

    assert isinstance(symbol, Wildcard)

    if symbol is ParameterWildcard:
        return tuple(get_pk_parameters(model))

    assert symbol is CovariateWildcard

    return tuple(column.name for column in model.datainfo if column.type == 'covariate')


def _interpret_ref(model, symbol):
    if symbol.name in ['ABSORPTION', 'ELIMINATION', 'DISTRIBUTION']:
        return tuple(get_pk_parameters(model, kind=symbol.name.lower()))
    elif symbol.name == 'CATEGORICAL':
        return tuple(
            column.name
            for column in model.datainfo
            if column.type == 'covariate' and not column.continuous
        )
    elif symbol.name == 'CONTINUOUS':
        return tuple(
            column.name
            for column in model.datainfo
            if column.type == 'covariate' and column.continuous
        )
    elif symbol.name == 'IIV':
        return tuple(get_individual_parameters(model, level='iiv'))
    elif symbol.name == 'PD':
        return tuple(get_pd_parameters(model))
    elif symbol.name == 'PD_IIV':
        return [
            pd_param
            for pd_param in get_pd_parameters(model)
            if len(get_parameter_rv(model, pd_param)) > 0
        ]
    elif symbol.name == 'PK':
        return tuple(get_pk_parameters(model))
    elif symbol.name == 'BIOAVAIL':
        return tuple(_get_bioaval_parameters(model))
    elif symbol.name == 'PK_IIV':
        return [
            pk_param
            for pk_param in get_pk_parameters(model)
            if len(get_parameter_rv(model, pk_param)) > 0
        ]
    else:
        return ()


def _get_bioaval_parameters(model):
    """Find all bioavail individual parameters
    Handle the case where one statement is only a logistic transformation
    """
    all_bio = get_bioavailability(model)
    pk = model.statements.before_odes
    found = []
    for _, bio in all_bio.items():
        if isinstance(bio, sympy.Symbol):
            ass = pk.find_assignment(bio)
            rhs_symbs = ass.rhs_symbols
            if len(rhs_symbs) == 1:
                symb = rhs_symbs.pop()
                ass2 = pk.find_assignment(symb)
                found.append(ass2.symbol.name)
            else:
                found += [s.name for s in rhs_symbs]
    return found



from __future__ import annotations

import os
from dataclasses import replace
from typing import TypeVar

from pharmpy.model import Model

from .context import insert_context
from .results import ModelfitResults, Results
from .workflow import Workflow, WorkflowBuilder

T = TypeVar('T')


def execute_workflow(
    workflow: Workflow[T], dispatcher=None, database=None, path=None, resume=False
) -> T:
    """Execute workflow

    Parameters
    ----------
    workflow : Workflow
        Workflow to execute
    dispatcher : ExecutionDispatcher
        Dispatcher to use
    database : ToolDatabase
        Tool database to use. None for the default Tool database.
    path : Path
        Path to use for database if applicable.
    resume : bool
        Whether to allow resuming previous workflow execution.

    Returns
    -------
    Results
        Results object created by workflow
    """
    # FIXME: Return type is not always Results
    if dispatcher is None:
        from pharmpy.workflows import default_dispatcher

        dispatcher = default_dispatcher
    if database is None:
        from pharmpy.workflows import default_tool_database

        database = default_tool_database(
            toolname=workflow.name, path=path, exist_ok=resume
        )  # TODO: database -> tool_database

    # For all input models set new database and read in results
    original_input_models = []
    input_models = []
    wb = WorkflowBuilder(workflow)
    for task in workflow.tasks:
        new_inp = []

        for inp in task.task_input:
            if isinstance(inp, Model):
                original_input_models.append(inp)
                inp.modelfit_results  # To read in the results
                new_model = inp.replace(parent_model=inp.name)
                new_inp.append(new_model)
                input_models.append(new_model)
            else:
                new_inp.append(inp)

        new_task = task.replace(task_input=tuple(new_inp))
        wb.replace_task(task, new_task)

    insert_context(wb, database)
    workflow = Workflow(wb)

    res: T = dispatcher.run(workflow)

    if isinstance(res, Results) and not isinstance(res, ModelfitResults):
        if hasattr(res, 'tool_database'):
            res = replace(res, tool_database=database)
        database.store_results(res)
        if hasattr(res, 'rst_path'):
            from pharmpy.tools.reporting import create_report

            if os.name == 'nt':
                # Workaround for issue with dask versions >= 2023.7.0 on Windows
                import asyncio

                asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
            create_report(res, database.path)
    elif isinstance(res, Model):
        # original_input_models[0].modelfit_results = res.modelfit_results
        pass
    elif isinstance(res, list) or isinstance(res, tuple):
        # index = {model.name: model for model in res}
        # for original_model in original_input_models:
        #    if (model := index.get(original_model.name, None)) is not None:
        #        original_model.modelfit_results = model.modelfit_results
        pass

    return res



def name_mangle(s: str) -> str:
    """
    Changes the format of parameter name to avoid using parenthesis

    Parameters
    ----------
    s : str
        Parameter name to be changed

    Returns
    -------
    str
        Parameter name with parenthesis removed

    Example
    -------
    name_mangle("ETA(1)")
    -> "ETA1"

    """
    return s.replace('(', '').replace(')', '').replace(',', '_')



import os.path
from pathlib import Path

import pytest
import sympy

from pharmpy.model import Model as BaseModel
from pharmpy.model.external.nlmixr.model import Model as nlmixrModel
from pharmpy.model.external.nonmem import Model as NMModel
from pharmpy.modeling import (
    convert_model,
    create_basic_pk_model,
    create_joint_distribution,
    filter_dataset,
    fix_parameters,
    get_config_path,
    get_model_code,
    get_model_covariates,
    load_example_model,
    read_model,
    read_model_from_string,
    remove_unused_parameters_and_rvs,
    set_name,
    write_model,
)
from pharmpy.tools import read_modelfit_results


def test_get_config_path():
    with pytest.warns(UserWarning):
        assert get_config_path() is None


def test_read_model_path(testdata):
    path = testdata / 'nonmem' / 'minimal.mod'
    model = read_model(path)
    assert model.parameters['THETA_1'].init == 0.1


def test_read_model_str(testdata):
    path = str(testdata / 'nonmem' / 'minimal.mod')
    model = read_model(path)
    assert model.parameters['THETA_1'].init == 0.1


def test_read_model_expanduser(testdata):
    model_path = testdata / "nonmem" / "minimal.mod"
    model_path_relative_to_home = ''
    try:
        model_path_relative_to_home = model_path.relative_to(Path.home())
    except ValueError:
        pytest.skip(f'{model_path} is not a descendant of home directory ({Path.home()})')
    path = os.path.join('~', model_path_relative_to_home)
    model = read_model(path)
    assert model.parameters['THETA_1'].init == 0.1


def test_read_model_from_string():
    model = read_model_from_string(
        """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    )
    assert model.parameters['THETA_1'].init == 0.1


# NOTE: Will warn on GHA for Windows due to different drives
@pytest.mark.filterwarnings("ignore::UserWarning")
def test_write_model(testdata, load_model_for_test, tmp_path):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    write_model(model, tmp_path / 'run1.mod')
    assert Path(tmp_path / 'run1.mod').is_file()


def test_generate_model_code(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = fix_parameters(model, ['THETA_1'])
    assert get_model_code(model).split('\n')[7] == '$THETA 0.1 FIX'


def test_load_example_model():
    model = load_example_model("pheno")
    assert len(model.parameters) == 6

    model = load_example_model("moxo")
    assert len(model.parameters) == 11

    with pytest.raises(ValueError):
        load_example_model("grekztalb23=")


def test_get_model_covariates(pheno, testdata, load_model_for_test):
    assert set(get_model_covariates(pheno)) == {
        sympy.Symbol('APGR'),
        sympy.Symbol('WGT'),
    }
    minimal = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    assert set(get_model_covariates(minimal)) == set()


def test_set_name(pheno):
    model = set_name(pheno, "run1")
    assert model.name == "run1"


def test_convert_model():
    model = load_example_model("pheno")

    run1 = convert_model(model, "nlmixr")
    run2 = convert_model(run1, "nonmem")

    assert model.name == run2.name == run1.name
    assert model.parameters == run1.parameters == run2.parameters
    assert model.statements == run1.statements == run2.statements
    assert isinstance(run1, nlmixrModel)
    assert isinstance(run2, NMModel)

    base = create_basic_pk_model('iv')

    assert isinstance(base, BaseModel)

    run3 = convert_model(base, 'nonmem')

    assert base.name == run3.name
    # Only checking parameters due to NONMEM parametrizations
    assert base.parameters == run3.parameters
    assert isinstance(run3, NMModel)


def test_remove_unused_parameters_and_rvs(load_model_for_test, pheno_path):
    pheno = load_model_for_test(pheno_path)
    res = read_modelfit_results(pheno_path)
    model = remove_unused_parameters_and_rvs(pheno)
    model = create_joint_distribution(model, individual_estimates=res.individual_estimates)
    statements = model.statements
    i = statements.index(statements.find_assignment('CL'))
    model = model.replace(statements=model.statements[0:i] + model.statements[i + 1 :])
    model = remove_unused_parameters_and_rvs(model)
    assert len(model.random_variables['ETA_2'].names) == 1


def test_filter_dataset(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem/pheno_pd.mod')
    assert 2 in model.dataset['DVID'].values
    model_filtered = filter_dataset(model, 'DVID == 1')
    assert 2 not in model_filtered.dataset['DVID'].values



"""
:meta private:
"""
from pharmpy.deps import sympy
from pharmpy.model import (
    Assignment,
    Compartment,
    CompartmentalSystem,
    CompartmentalSystemBuilder,
    Model,
    output,
)

from .error import set_proportional_error_model
from .odes import _find_noncov_theta, add_individual_parameter, set_first_order_absorption
from .parameters import set_initial_estimates, set_lower_bounds, set_upper_bounds


def add_metabolite(model: Model, drug_dvid: int = 1, presystemic: bool = False):
    """Adds a metabolite compartment to a model

    The flow from the central compartment to the metabolite compartment
    will be unidirectional.

    Presystemic indicate that the metabolite compartment will be
    directly connected to the DEPOT. If a depot compartment is not present,
    one will be created.

    Parameters
    ----------
    model : Model
        Pharmpy model
    drug_dvid : int
        DVID for drug (assuming all other DVIDs being for metabolites)
    presystemic : bool
        Decide wether or not to add metabolite as a presystemetic fixed drug.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = add_metabolite(model)

    """

    if presystemic:
        depot = model.statements.ode_system.find_depot(model.statements)
        if not depot:
            model = set_first_order_absorption(model)
            depot = model.statements.ode_system.find_depot(model.statements)

    # TODO: Implement possibility of converting plain metabolite to presystemic

    clm = sympy.Symbol('CLM')
    model = add_individual_parameter(model, clm.name)
    vm = sympy.Symbol('VM')
    model = add_individual_parameter(model, vm.name)

    odes = model.statements.ode_system
    central = odes.central_compartment
    ke = odes.get_flow(central, output)
    cl, vc = ke.as_numer_denom()

    if vc != 1:
        pop_cl = _find_noncov_theta(model, cl)
        pop_vc = _find_noncov_theta(model, vc)
        pop_clm_init = model.parameters[pop_cl].init * 2.0
        pop_vm_init = model.parameters[pop_vc].init * 0.5
        model = set_initial_estimates(model, {'POP_CLM': pop_clm_init, 'POP_VM': pop_vm_init})

    cb = CompartmentalSystemBuilder(odes)
    metacomp = Compartment.create(name="METABOLITE")
    cb.add_compartment(metacomp)
    cb.add_flow(metacomp, output, clm / vm)

    cb.add_flow(central, metacomp, ke)
    cb.remove_flow(central, output)
    if presystemic:
        fpre = sympy.Symbol('FPRE')
        model = add_individual_parameter(model, fpre.name)
        model = set_lower_bounds(model, {'POP_FPRE': sympy.Number(0)})
        model = set_upper_bounds(model, {'POP_FPRE': sympy.Number(1)})
        ka = odes.get_flow(depot, central)
        cb.add_flow(depot, metacomp, fpre * ka)
        cb.remove_flow(depot, central)
        cb.add_flow(depot, central, ka * (1 - fpre))

    # FIXME: drug_dvid is never used, use it here?
    # dvid_col = model.datainfo.typeix['dvid'][0]
    # dvids = dvid_col.categories

    amount = metacomp.amount

    if presystemic:
        # QUESTION: Add bioavailability to depot?
        cb.set_bioavailability(depot, 1 / (1 - fpre))
        model = model.replace(
            statements=model.statements.before_odes
            + CompartmentalSystem(cb)
            + model.statements.after_odes
        )
        model = model.update_source()
        cs = model.statements.ode_system
        bio = model.statements.ode_system.find_compartment(depot.name).bioavailability
        conc = Assignment(sympy.Symbol('CONC_M'), amount / vm / bio)
    else:
        cs = CompartmentalSystem(cb)
        conc = Assignment(sympy.Symbol('CONC_M'), amount / vm)
    y_m = sympy.Symbol('Y_M')
    y = Assignment(y_m, conc.symbol)
    original_y = next(iter(model.dependent_variables))
    ind = model.statements.after_odes.find_assignment_index(original_y)
    old_after = model.statements.after_odes
    new_after = old_after[: ind + 1] + y + old_after[ind + 1 :]
    error = conc + new_after
    dvs = model.dependent_variables.replace(y_m, 2)  # FIXME: Should be next DVID in categories
    statements = model.statements.before_odes + cs + error
    model = model.replace(statements=statements, dependent_variables=dvs)
    model = set_proportional_error_model(model, dv=2, zero_protection=False)
    return model


def has_presystemic_metabolite(model: Model):
    """Checks whether a model has a presystemic metabolite

    If pre-systemic drug there will be a flow from DEPOT to METABOLITE as well
    as being a flow from the CENTRAL to METABOLITE

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    bool
        Whether a model has presystemic metabolite

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = add_metabolite(model, presystemic=True)
    >>> has_presystemic_metabolite(model)
    True

    """
    odes = model.statements.ode_system
    central = odes.central_compartment
    metabolite = odes.find_compartment("METABOLITE")
    depot = odes.find_depot(model.statements)

    CM = odes.get_flow(central, metabolite)
    DM = odes.get_flow(depot, metabolite)

    if CM != sympy.Number(0) and DM != sympy.Number(0):
        return True
    else:
        return False



import re

from pharmpy.model import ModelSyntaxError

from .abbreviated_record import AbbreviatedRecord
from .code_record import CodeRecord
from .data_record import DataRecord
from .estimation_record import EstimationRecord
from .etas_record import EtasRecord
from .model_record import ModelRecord
from .omega_record import OmegaRecord
from .option_record import OptionRecord
from .parsers import (
    AbbreviatedRecordParser,
    CodeRecordParser,
    DataRecordParser,
    OmegaRecordParser,
    OptionRecordParser,
    ProblemRecordParser,
    SimulationRecordParser,
    ThetaRecordParser,
)
from .problem_record import ProblemRecord
from .raw_record import RawRecord
from .simulation_record import SimulationRecord
from .sizes_record import SizesRecord
from .subroutine_record import SubroutineRecord
from .table_record import TableRecord
from .theta_record import ThetaRecord

# Dictionary from canonical record name to record class and non_empty rules of parser
known_records = {
    'ABBREVIATED': (AbbreviatedRecord, AbbreviatedRecordParser),
    'COVARIANCE': (OptionRecord, OptionRecordParser),
    'DATA': (DataRecord, DataRecordParser),
    'DES': (CodeRecord, CodeRecordParser),
    'ERROR': (CodeRecord, CodeRecordParser),
    'ESTIMATION': (EstimationRecord, OptionRecordParser),
    'ETAS': (EtasRecord, OptionRecordParser),
    'INPUT': (OptionRecord, OptionRecordParser),
    'MODEL': (ModelRecord, OptionRecordParser),
    'OMEGA': (OmegaRecord, OmegaRecordParser),
    'PK': (CodeRecord, CodeRecordParser),
    'PRED': (CodeRecord, CodeRecordParser),
    'PROBLEM': (ProblemRecord, ProblemRecordParser),
    'SIGMA': (OmegaRecord, OmegaRecordParser),
    'SIMULATION': (SimulationRecord, SimulationRecordParser),
    'SIZES': (SizesRecord, OptionRecordParser),
    'SUBROUTINES': (SubroutineRecord, OptionRecordParser),
    'TABLE': (TableRecord, OptionRecordParser),
    'THETA': (ThetaRecord, ThetaRecordParser),
}


def split_raw_record_name(line: str):
    """Splits the raw record name of the first line of a record from the rest of the record"""
    m = re.match(r'(\s*\$[A-za-z]+)(.*)', line, flags=re.MULTILINE | re.DOTALL)
    if m:
        return m.group(1, 2)
    else:
        raise ModelSyntaxError(f'Bad record name in: {line}')


def get_canonical_record_name(raw_name):
    """Gets the canonical (standardized) record name from a raw_name"""
    bare = raw_name.lstrip()[1:].upper()  # remove initial white space and the '$'
    if len(bare) >= 3:
        for name in known_records:
            if name.startswith(bare):
                return name
        # Synonyms
        if 'INFILE'.startswith(bare):
            return 'DATA'
        elif 'SUBS'.startswith(bare):
            return 'SUBROUTINES'
        elif bare == 'SIML' or bare == 'SIMULATE':
            return 'SIMULATION'
        elif bare == 'COVR':
            return 'COVARIANCE'
        elif bare == 'ESTM':
            return 'ESTIMATION'
    elif bare == 'PK':
        return bare
    return None


def create_record(chunk: str):
    raw_name, content = split_raw_record_name(chunk)
    canonical_name = get_canonical_record_name(raw_name)
    if canonical_name:
        name = canonical_name
        record_class, record_parser_class = known_records[name]
        root = record_parser_class(content).root
        record = record_class(name, raw_name, root)
    else:
        name = raw_name[1:]
        record = RawRecord(content, name, raw_name)

    return record



"""Pretty printing of :class:`generic.AttrTree` via :class:`treeprint.Node`."""
from __future__ import annotations

from functools import partial
from typing import Literal, Union

from .tree import Leaf, Tree
from .treeprint import Node

MAXLEN = 60


def _preview(content: str):
    """Represents str and cut into shorts if too long"""
    cut = len(content) - MAXLEN
    if cut > 0:
        preview = '"' + repr(content[0:MAXLEN])[1:-1] + '"'
        preview = '%s+%d' % (preview, cut)
    else:
        preview = '"' + repr(content)[1:-1] + '"'
    return preview


def _formatter(content: Union[bool, Literal['full']], ast_node):
    """Formats AST node (Tree and Leaf) for treeprint"""
    if content:
        if content == 'full':
            lines = [ast_node.rule] + str(ast_node).splitlines(True)
            lines[1:] = [_preview(line) for line in lines[1:]]
            return '\n'.join(lines)
        else:
            return '%s %s' % (ast_node.rule, _preview(str(ast_node)))
    else:
        return '%s' % (ast_node.rule,)


def _format_tree(content: Union[bool, Literal['full']], ast_tree: Tree):
    return _formatter(content, ast_tree)


def _format_token(content: Union[bool, Literal['full']], ast_token: Leaf):
    return _formatter(content, ast_token)


def transform(ast_tree_or_token: Union[Tree, Leaf], content: Union[bool, Literal['full']] = True):
    """
    Traverses tree and generates :class:`treeprint.Node`, which can format a
    multiline (command) tree-styled string.

    Args:
        content: If True, include text preview (or all if 'full') on each node.

    Returns:
        Multiline string, ready for the printing press.
    """

    if isinstance(ast_tree_or_token, Leaf):
        formatter = partial(_format_token, content)
        return Node(ast_tree_or_token, cls_str=formatter, children=())
    elif isinstance(ast_tree_or_token, Tree):
        nodes = ast_tree_or_token.children
        if isinstance(nodes, str):
            raise TypeError(
                "'children' of tree appears to be 'str' (expects list/iterable): %s"
                % repr(ast_tree_or_token)
            )
        formatter = partial(_format_tree, content)
        children = tuple(transform(ast_node, content) for ast_node in list(nodes))
        return Node(ast_tree_or_token, cls_str=formatter, children=children)
    else:
        raise TypeError(
            "can't transform %s object (is not a Lark Tree or Token')"
            % repr(ast_tree_or_token.__class__.__name__)
        )



import shutil
from dataclasses import replace

import pytest

from pharmpy.deps import numpy as np
from pharmpy.deps import pandas as pd
from pharmpy.deps import sympy
from pharmpy.internals.fs.cwd import chdir
from pharmpy.modeling import (
    add_iiv,
    add_population_parameter,
    create_joint_distribution,
    fix_or_unfix_parameters,
    fix_parameters,
    fix_parameters_to,
    get_omegas,
    get_sigmas,
    get_thetas,
    set_lower_bounds,
    set_upper_bounds,
    unconstrain_parameters,
    unfix_parameters,
    unfix_parameters_to,
    update_inits,
)
from pharmpy.tools import read_modelfit_results


def test_get_thetas(pheno):
    thetas = get_thetas(pheno)
    assert len(thetas) == 3
    assert thetas['PTVCL'].init == 0.00469307


def test_get_omegas(pheno):
    omegas = get_omegas(pheno)
    assert len(omegas) == 2
    assert omegas['IVCL'].init == 0.0309626


def test_get_sigmas(pheno):
    sigmas = get_sigmas(pheno)
    assert len(sigmas) == 1
    assert sigmas['SIGMA_1_1'].init == 0.013241


def test_fix_parameters(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    assert not model.parameters['THETA_1'].fix
    model = fix_parameters(model, ['THETA_1'])
    assert model.parameters['THETA_1'].fix

    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    assert not model.parameters['THETA_1'].fix
    model = fix_parameters(model, 'THETA_1')
    assert model.parameters['THETA_1'].fix


def test_unfix_parameters(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = fix_parameters(model, ['THETA_1'])
    assert model.parameters['THETA_1'].fix
    model = unfix_parameters(model, ['THETA_1'])
    assert not model.parameters['THETA_1'].fix

    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = fix_parameters(model, 'THETA_1')
    assert model.parameters['THETA_1'].fix
    model = unfix_parameters(model, 'THETA_1')
    assert not model.parameters['THETA_1'].fix


def test_fix_or_unfix_parameters(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = fix_or_unfix_parameters(model, {'THETA_1': True})
    assert model.parameters['THETA_1'].fix
    model = fix_or_unfix_parameters(model, {'THETA_1': False})
    assert not model.parameters['THETA_1'].fix


def test_unconstrain_parameters(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    fix_or_unfix_parameters(model, {'THETA_1': True})
    unconstrain_parameters(model, ['THETA_1'])
    assert not model.parameters['THETA_1'].fix
    fix_or_unfix_parameters(model, {'THETA_1': True})
    unconstrain_parameters(model, 'THETA_1')
    assert not model.parameters['THETA_1'].fix


def test_fix_parameters_to(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = fix_parameters_to(model, {'THETA_1': 0})
    assert model.parameters['THETA_1'].fix
    assert model.parameters['THETA_1'].init == 0

    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = fix_parameters_to(model, {'THETA_1': 0, 'OMEGA_1_1': 0})
    assert model.parameters['THETA_1'].fix
    assert model.parameters['THETA_1'].init == 0
    assert model.parameters['THETA_1'].fix
    assert model.parameters['OMEGA_1_1'].init == 0

    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = fix_parameters_to(model, {'THETA_1': 0, 'OMEGA_1_1': 1})
    assert model.parameters['THETA_1'].init == 0
    assert model.parameters['OMEGA_1_1'].init == 1


def test_unfix_parameters_to(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = fix_parameters(model, ['THETA_1'])
    assert model.parameters['THETA_1'].fix
    model = unfix_parameters_to(model, {'THETA_1': 0})
    assert not model.parameters['THETA_1'].fix
    assert model.parameters['THETA_1'].init == 0

    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = fix_parameters(model, ['THETA_1', 'OMEGA_1_1'])
    assert model.parameters['THETA_1'].fix
    assert model.parameters['OMEGA_1_1'].fix
    model = unfix_parameters_to(model, {'THETA_1': 0, 'OMEGA_1_1': 0})
    assert not model.parameters['THETA_1'].fix
    assert not model.parameters['OMEGA_1_1'].fix
    assert model.parameters['THETA_1'].init == 0
    assert model.parameters['OMEGA_1_1'].init == 0


def test_set_bounds(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = set_upper_bounds(model, {'THETA_1': 100})
    assert model.parameters['THETA_1'].upper == 100
    assert model.parameters['OMEGA_1_1'].upper == sympy.oo
    model = set_lower_bounds(model, {'THETA_1': -100})
    assert model.parameters['THETA_1'].lower == -100
    assert model.parameters['OMEGA_1_1'].lower == 0


def test_add_population_parameter(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model = add_population_parameter(model, 'NEWPARAM', 23)
    assert len(model.parameters) == 4
    assert model.parameters['NEWPARAM'].init == 23


def test_update_inits_move_est(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    res = read_modelfit_results(pheno_path)

    model = create_joint_distribution(model, individual_estimates=res.individual_estimates)
    model = add_iiv(model, 'S1', 'add')

    param_est = res.parameter_estimates.copy()
    param_est['IIV_CL_IIV_V'] = 0.0285  # Correlation > 0.99
    param_est['IIV_S1'] = 0.0005

    model = update_inits(model, param_est, move_est_close_to_bounds=True)

    assert model.parameters['IVCL'].init == param_est['IVCL']
    assert model.parameters['IIV_S1'].init == 0.01
    assert round(model.parameters['IIV_CL_IIV_V'].init, 6) == 0.025757


def test_update_inits_zero_fix(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    d = {name: 0 for name in model.random_variables.iiv.parameter_names}
    model = fix_parameters_to(model, d)
    res = read_modelfit_results(pheno_path)
    param_est = res.parameter_estimates.drop(index=['IVCL'])
    model = update_inits(model, param_est)
    assert model.parameters['IVCL'].init == 0
    assert model.parameters['IVCL'].fix

    model = load_model_for_test(pheno_path)
    d = {name: 0 for name in model.random_variables.iiv.parameter_names}
    model = fix_parameters_to(model, d)
    param_est = res.parameter_estimates.drop(index=['IVCL'])
    model = update_inits(model, param_est, move_est_close_to_bounds=True)
    assert model.parameters['IVCL'].init == 0
    assert model.parameters['IVCL'].fix


def test_update_inits_no_res(load_model_for_test, testdata, tmp_path):
    shutil.copy(testdata / 'nonmem/pheno.mod', tmp_path / 'run1.mod')
    shutil.copy(testdata / 'nonmem/pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        shutil.copy(testdata / 'nonmem/pheno.ext', tmp_path / 'run1.ext')
        shutil.copy(testdata / 'nonmem/pheno.lst', tmp_path / 'run1.lst')

        model = load_model_for_test('run1.mod')
        res = read_modelfit_results('run1.mod')

        modelfit_results = replace(
            res,
            parameter_estimates=pd.Series(
                np.nan, name='estimates', index=list(model.parameters.nonfixed.inits.keys())
            ),
        )

        with pytest.raises(ValueError):
            update_inits(model, modelfit_results.parameter_estimates)



from typing import Optional, Union

from pharmpy.deps import pandas as pd
from pharmpy.model import Model
from pharmpy.modeling import (
    add_iiv,
    fix_parameters_to,
    set_baseline_effect,
    set_initial_estimates,
    set_lower_bounds,
    set_name,
    unconstrain_parameters,
)
from pharmpy.tools.mfl.helpers import funcs, structsearch_pd_features

from ..mfl.parse import parse as mfl_parse


def create_baseline_pd_model(model: Model, ests: pd.Series, b_init: Optional[float] = None):
    """Create baseline pkpd model

    Parameters
    ----------
    model : Model
        Pharmpy PK model
    ests : pd.Series
       List of estimated PK parameters
    b_init : float
        Initial estimate for baseline

    Returns
    -------
    Baseline PD model
    """
    baseline_model = set_baseline_effect(model, expr='const')
    baseline_model = set_name(baseline_model, "baseline_model")
    baseline_model = baseline_model.replace(description="baseline_model")
    baseline_model = fix_parameters_to(baseline_model, ests)
    baseline_model = add_iiv(baseline_model, ["B"], "exp")
    if b_init is not None:
        baseline_model = set_initial_estimates(baseline_model, {'POP_B': b_init})
    return baseline_model


def create_pkpd_models(
    model: Model,
    search_space: str,
    b_init: Optional[Union[int, float]] = None,
    ests: Optional[pd.Series] = None,
    emax_init: Optional[Union[int, float]] = None,
    ec50_init: Optional[Union[int, float]] = None,
    met_init: Optional[Union[int, float]] = None,
):
    """Create pkpd models

    Parameters
    ----------
    model : Model
        Pharmpy PK model
    search_space : str
        search space for pkpd models
    b_init : float
       Initial estimate for baseline
    ests : pd.Series
       List of estimated PK parameters
    emax_init : float
        Initial estimate for E_max
    ec50_init : float
        Initial estimate for EC_50
    met_init : float
        Initial estimate for MET (mean equilibration time)

    Returns
    -------
    List of pharmpy models
    """
    mfl_statements = mfl_parse(search_space)
    functions = funcs(model, mfl_statements, structsearch_pd_features)

    models = []
    for key, func in functions.items():
        pkpd_model = func(model)
        description = '_'.join(key)
        pkpd_model = pkpd_model.replace(description=description)
        models.append(pkpd_model)

    final_models = []
    for index, pkpd_model in enumerate(models, 1):
        pkpd_model = set_name(pkpd_model, f"structsearch_run{index}")

        # Initial values
        if b_init is not None:
            pkpd_model = set_initial_estimates(pkpd_model, {'POP_B': b_init})
        if ests is not None:
            pkpd_model = fix_parameters_to(pkpd_model, ests)
        if emax_init is not None:
            pkpd_model = set_initial_estimates(pkpd_model, {'POP_E_MAX': emax_init})
        if ec50_init is not None:
            pkpd_model = set_initial_estimates(pkpd_model, {'POP_EC_50': ec50_init})
        if emax_init is not None and ec50_init is not None:
            pkpd_model = set_initial_estimates(pkpd_model, {'POP_SLOPE': emax_init / ec50_init})
        if met_init is not None:
            pkpd_model = set_initial_estimates(pkpd_model, {'POP_MET': met_init})

        pkpd_model = unconstrain_parameters(pkpd_model, ['POP_SLOPE'])
        pkpd_model = set_lower_bounds(pkpd_model, {'POP_E_MAX': -1.0})

        # Set iiv
        for parameter in ["E_MAX", "SLOPE"]:
            try:
                pkpd_model = add_iiv(pkpd_model, [parameter], "prop", initial_estimate=0.1)
            except ValueError:
                pass
        try:
            pkpd_model = add_iiv(pkpd_model, 'B', "exp", initial_estimate=0.1)
        except ValueError:
            pass

        final_models.append(pkpd_model)

    return final_models


def create_pk_model(model: Model):
    # Create copy of model with filtered dataset.
    # FIXME: This function needs to be removed later
    pk_dataset = model.dataset
    pk_dataset = pk_dataset[pk_dataset["DVID"] != 2]
    pk_model = model.replace(
        dataset=pk_dataset,
        description=model.description + ". Removed rows with DVID=2.",
        name="PK_" + model.name,
    )
    return pk_model



import json
import os
import os.path
import shutil
from pathlib import Path

import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.modeling import add_time_after_dose
from pharmpy.tools import read_modelfit_results
from pharmpy.workflows import (
    LocalDirectoryDatabase,
    LocalModelDirectoryDatabase,
    ModelDatabase,
    ModelEntry,
    NullModelDatabase,
)


def test_base_class():
    with pytest.raises(TypeError):
        ModelDatabase()


def test_local_directory(tmp_path):
    with chdir(tmp_path):
        os.mkdir("database")
        db = LocalDirectoryDatabase("database")
        with open("file.txt", "w") as fh:
            print("Hello!", file=fh)
        db.store_local_file(None, "file.txt")
        with open("database/file.txt", "r") as fh:
            assert fh.read() == "Hello!\n"

        dirname = "doesnotexist"
        db = LocalDirectoryDatabase(dirname)
        assert os.path.isdir(dirname)


def test_null_database():
    db = NullModelDatabase(klr=123, f="oe")
    db.store_local_file("path", 34)


def test_store_model(tmp_path, load_model_for_test, testdata):
    sep = os.path.sep
    with chdir(tmp_path):
        datadir = testdata / 'nonmem'
        shutil.copy(datadir / 'pheno_real.mod', 'pheno_real.mod')
        shutil.copy(datadir / 'pheno.dta', 'pheno.dta')
        model = load_model_for_test("pheno_real.mod")

        db = LocalModelDirectoryDatabase("database")
        db.store_model(model)

        with open("database/.datasets/pheno_real.csv", "r") as fh:
            line = fh.readline()
            assert line == "ID,TIME,AMT,WGT,APGR,DV,FA1,FA2\n"

        with open("database/.datasets/pheno_real.datainfo", "r") as fh:
            obj = json.load(fh)
            assert 'columns' in obj
            assert obj['path'] == 'pheno_real.csv'

        with open("database/pheno_real/pheno_real.mod", "r") as fh:
            line = fh.readline()
            assert line == "$PROBLEM PHENOBARB SIMPLE MODEL\n"
            line = fh.readline()
            assert line == f'$DATA ..{sep}.datasets{sep}pheno_real.csv IGNORE=@\n'

        run1 = model.replace(name="run1")
        db.store_model(run1)

        assert not (Path("database") / ".datasets" / "run1.csv").is_file()

        with open("database/run1/run1.mod", "r") as fh:
            line = fh.readline()
            assert line == "$PROBLEM PHENOBARB SIMPLE MODEL\n"
            line = fh.readline()
            assert line == f'$DATA ..{sep}.datasets{sep}pheno_real.csv IGNORE=@\n'

        run2 = model.replace(name="run2")
        run2 = add_time_after_dose(run2)
        db.store_model(run2)

        with open("database/.datasets/run2.csv", "r") as fh:
            line = fh.readline()
            assert line == "ID,TIME,AMT,WGT,APGR,DV,FA1,FA2,TAD\n"

        with open("database/.datasets/run2.datainfo", "r") as fh:
            obj = json.load(fh)
            assert 'columns' in obj
            assert obj['path'] == 'run2.csv'

        with open("database/run2/run2.mod", "r") as fh:
            line = fh.readline()
            assert line == "$PROBLEM PHENOBARB SIMPLE MODEL\n"
            line = fh.readline()
            assert line == f'$DATA ..{sep}.datasets{sep}run2.csv IGNORE=@\n'


def test_store_and_retrieve_model_entry(tmp_path, load_model_for_test, testdata):
    sep = os.path.sep
    with chdir(tmp_path):
        datadir = testdata / 'nonmem'
        for path in (testdata / 'nonmem').glob('pheno_real.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy(datadir / 'pheno.dta', 'pheno.dta')
        model = load_model_for_test("pheno_real.mod")
        modelfit_results = read_modelfit_results("pheno_real.mod")
        model_entry = ModelEntry(
            model=model,
            modelfit_results=modelfit_results,
            parent=model.parent_model,
            log=modelfit_results.log,
        )

        db = LocalModelDirectoryDatabase("database")
        db.store_model_entry(model_entry)

        with open("database/.datasets/pheno_real.csv", "r") as fh:
            line = fh.readline()
            assert line == "ID,TIME,AMT,WGT,APGR,DV,FA1,FA2\n"

        with open("database/.datasets/pheno_real.datainfo", "r") as fh:
            obj = json.load(fh)
            assert 'columns' in obj
            assert obj['path'] == 'pheno_real.csv'

        with open("database/pheno_real/pheno_real.mod", "r") as fh:
            line = fh.readline()
            assert line == "$PROBLEM PHENOBARB SIMPLE MODEL\n"
            line = fh.readline()
            assert line == f'$DATA ..{sep}.datasets{sep}pheno_real.csv IGNORE=@\n'

        model_entry_retrieve = db.retrieve_model_entry(model.name)

        assert model_entry_retrieve.model == model
        assert model_entry_retrieve.modelfit_results.name == modelfit_results.name



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

from pharmpy.deps import pandas as pd
from pharmpy.model import Model
from pharmpy.modeling import plot_individual_predictions
from pharmpy.tools import read_modelfit_results
from pharmpy.tools.simfit.results import psn_simfit_results
from pharmpy.workflows import Results


@dataclass(frozen=True)
class SimevalResults(Results):
    """Simeval results class"""

    sampled_iofv: Optional[Any] = None
    iofv_summary: Optional[Any] = None
    individual_predictions_plot: Optional[Any] = None


def calculate_results(original_model, original_results, simfit_results):
    """Calculate simeval results"""
    sampled_iofv = pd.concat(
        [res.individual_ofv for res in simfit_results.modelfit_results],
        axis=1,
        keys=range(1, len(simfit_results.modelfit_results) + 1),
    )
    origiofv = original_results.individual_ofv
    iofv_summary = pd.DataFrame(
        {
            'original': origiofv,
            'sampled_mean': sampled_iofv.T.mean(),
            'sampled_stdev': sampled_iofv.T.std(),
        }
    )
    iofv_summary['residual'] = (
        iofv_summary['original'] - iofv_summary['sampled_mean']
    ) / iofv_summary['sampled_stdev']
    iofv_summary['residual_q1'] = (
        iofv_summary['original'] - sampled_iofv.T.quantile(0.25)
    ) / iofv_summary['sampled_stdev']
    iofv_summary['residual_q3'] = (
        iofv_summary['original'] - sampled_iofv.T.quantile(0.75)
    ) / iofv_summary['sampled_stdev']
    iofv_summary['residual_outlier'] = iofv_summary['residual'] >= 3

    ids = iofv_summary.index[iofv_summary['residual_outlier']].tolist()
    id_plot = None
    if ids:
        try:
            id_plot = plot_individual_predictions(
                original_model,
                original_results.predictions[['IPRED', 'PRED']],
                individuals=ids,
            )
        except Exception:
            pass

    res = SimevalResults(
        sampled_iofv=sampled_iofv,
        iofv_summary=iofv_summary,
        individual_predictions_plot=id_plot,
    )
    return res


def psn_simeval_results(path):
    path = Path(path)
    simfit_paths = (path / 'm1').glob('sim-*.mod')
    simfit_results = psn_simfit_results(simfit_paths)
    original = Model.parse_model(path / 'm1' / 'original.mod')
    original_results = read_modelfit_results(path / 'm1' / 'original.mod')
    res = calculate_results(original, original_results, simfit_results)

    # Add CWRES outliers as 2 in data_flag
    # Reading PsN results for now
    # df = pd.read_csv(path / 'summary_cwres.csv')
    # outliers = df['OUTLIER'].fillna(0).astype(int)
    # outliers.replace({1.0: 2}, inplace=True)
    # outliers = outliers + res.data_flag
    # outliers.replace({3: 1}, inplace=True)
    # res.data_flag = outliers
    return res



import warnings
from pathlib import Path
from typing import Callable, List, Literal, Optional, Tuple, Union

from pharmpy.deps import numpy as np
from pharmpy.deps import pandas as pd
from pharmpy.deps import sympy
from pharmpy.internals.fn.type import check_list, with_runtime_arguments_type_check
from pharmpy.model import Model
from pharmpy.modeling import (
    add_parameter_uncertainty_step,
    has_mixed_mm_fo_elimination,
    plot_abs_cwres_vs_ipred,
    plot_cwres_vs_idv,
    plot_dv_vs_ipred,
    plot_dv_vs_pred,
)
from pharmpy.modeling.blq import has_blq_transformation, transform_blq
from pharmpy.modeling.common import convert_model, filter_dataset
from pharmpy.modeling.covariate_effect import get_covariates_allowed_in_covariate_effect
from pharmpy.modeling.parameter_variability import get_occasion_levels
from pharmpy.modeling.tmdd import DV_TYPES
from pharmpy.reporting import generate_report
from pharmpy.tools import retrieve_models, summarize_errors, write_results
from pharmpy.tools.allometry.tool import validate_allometric_variable
from pharmpy.tools.mfl.feature.covariate import covariates as extract_covariates
from pharmpy.tools.mfl.feature.covariate import spec as covariate_spec
from pharmpy.tools.mfl.filter import (
    COVSEARCH_STATEMENT_TYPES,
    STRUCTSEARCH_STATEMENT_TYPES,
    mfl_filtering,
)
from pharmpy.tools.mfl.parse import ModelFeatures, get_model_features
from pharmpy.tools.mfl.parse import parse as mfl_parse
from pharmpy.tools.mfl.statement.definition import Let
from pharmpy.tools.mfl.statement.feature.absorption import Absorption
from pharmpy.tools.mfl.statement.feature.covariate import Covariate, Ref
from pharmpy.tools.mfl.statement.feature.elimination import Elimination
from pharmpy.tools.mfl.statement.feature.lagtime import LagTime
from pharmpy.tools.mfl.statement.feature.peripherals import Peripherals
from pharmpy.tools.mfl.statement.feature.symbols import Name, Option, Wildcard
from pharmpy.tools.mfl.statement.feature.transits import Transits
from pharmpy.tools.mfl.statement.statement import Statement
from pharmpy.tools.mfl.stringify import stringify as mfl_stringify
from pharmpy.workflows import Results, default_tool_database
from pharmpy.workflows.results import ModelfitResults

from ..run import run_tool
from .results import AMDResults

ALLOWED_STRATEGY = ["all", "retries", "SIR", "SRI", "RSI"]
ALLOWED_ADMINISTRATION = ["iv", "oral", "ivoral"]
ALLOWED_MODELTYPE = ['basic_pk', 'pkpd', 'drug_metabolite', 'tmdd']
RETRIES_STRATEGIES = ["final", "all_final", "skip"]


def run_amd(
    input: Union[Model, Path, str],
    results: Optional[ModelfitResults] = None,
    modeltype: str = 'basic_pk',
    administration: str = 'oral',
    strategy: str = "all",
    cl_init: float = 0.01,
    vc_init: float = 1.0,
    mat_init: float = 0.1,
    b_init: Optional[Union[int, float]] = None,
    emax_init: Optional[Union[int, float]] = None,
    ec50_init: Optional[Union[int, float]] = None,
    met_init: Optional[Union[int, float]] = None,
    search_space: Optional[str] = None,
    lloq_method: Optional[str] = None,
    lloq_limit: Optional[str] = None,
    allometric_variable: Optional[Union[str, sympy.Symbol]] = None,
    occasion: Optional[str] = None,
    path: Optional[Union[str, Path]] = None,
    resume: bool = False,
    strictness: Optional[str] = "minimization_successful or (rounding_errors and sigdigs>=0.1)",
    dv_types: Optional[dict[Literal[DV_TYPES], int]] = None,
    mechanistic_covariates: Optional[List[str]] = None,
    retries_strategy: Literal["final", "all_final", "skip"] = "all_final",
    seed: Optional[Union[np.random.Generator, int]] = None,
    parameter_uncertainty_method: Optional[Literal['SANDWICH', 'CPG', 'OFIM']] = None,
    ignore_datainfo_fallback: bool = False,
):
    """Run Automatic Model Development (AMD) tool

    Runs structural modelsearch, IIV building, and ruvsearch

    Parameters
    ----------
    input : Model or Path
        Read model object/Path to a dataset
    results : ModelfitResults
        Reults of input if input is a model
    modeltype : str
        Type of model to build. Valid strings are 'basic_pk', 'pkpd', 'drug_metabolite' and 'tmdd'
    administration : str
        Route of administration. Either 'iv', 'oral' or 'ivoral'
    strategy : str
        Run algorithm for AMD procedure. Valid options are 'all', 'retries'. Default is all
    cl_init : float
        Initial estimate for the population clearance
    vc_init : float
        Initial estimate for the central compartment population volume
    mat_init : float
        Initial estimate for the mean absorption time (not for iv models)
    b_init : float
        Initial estimate for the baseline (PKPD model)
    emax_init : float
        Initial estimate for E_max (PKPD model)
    ec50_init : float
        Initial estimate for EC_50 (PKPD model)
    met_init : float
        Initial estimate for mean equilibration time (PKPD model)
    search_space : str
        MFL for search space for structural model
    lloq_method : str
        Method for how to remove LOQ data. See `transform_blq` for list of available methods
    lloq_limit : float
        Lower limit of quantification. If None LLOQ column from dataset will be used
    allometric_variable: str or Symbol
        Variable to use for allometry
    occasion : str
        Name of occasion column
    path : str or Path
        Path to run AMD in
    resume : bool
        Whether to allow resuming previous run
    strictness : str or None
        Strictness criteria
    dv_types : dict or None
        Dictionary of DV types for TMDD models with multiple DVs.
    mechanistic_covariates : list
        List of covariates to run in a separate proioritized covsearch run. The effects are extracted
        from the search space for covsearch
    retries_strategy: str
        Weither or not to run retries tool. Valid options are 'skip', 'all_final' or 'final'.
        Default is 'final'.
    seed : int or rng
        Random number generator or seed to be used.
    parameter_uncertainty_method: {'SANDWICH', 'CPG', 'OFIM'} or None
        Parameter uncertainty method.
    ignore_datainfo_fallback : bool
        Ignore using datainfo to get information not given by the user. Default is False

    Returns
    -------
    Model
        Reference to the same model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> from pharmpy.tools import run_amd, load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> run_amd(model, results=results)      # doctest: +SKIP

    See also
    --------
    run_iiv
    run_tool

    """
    args = locals()

    from pharmpy.model.external import nonmem  # FIXME: We should not depend on NONMEM

    if modeltype == 'pkpd':
        dv = 2
        iiv_strategy = 'pd_fullblock'
        # FIXME : DO THIS ONLY ONCE (DUPLICATED FURTHER DOWN)
        try:
            input_search_space_features = [] if search_space is None else mfl_parse(search_space)
        except:  # noqa E722
            raise ValueError(f'Invalid `search_space`, could not be parsed: "{search_space}"')

        structsearch_features = tuple(
            filter(
                lambda statement: isinstance(statement, STRUCTSEARCH_STATEMENT_TYPES),
                input_search_space_features,
            )
        )
        if not structsearch_features:
            structsearch_features = "DIRECTEFFECT(*);EFFECTCOMP(*);INDIRECTEFFECT(*,*)"
        else:
            structsearch_features = mfl_stringify(structsearch_features)
    else:
        dv = None
        iiv_strategy = 'fullblock'

    if type(input) is str:
        from pharmpy.modeling import create_basic_pk_model

        model = create_basic_pk_model(
            administration,
            dataset_path=input,
            cl_init=cl_init,
            vc_init=vc_init,
            mat_init=mat_init,
        )
        model = convert_model(model, 'nonmem')  # FIXME: Workaround for results retrieval system
    elif type(input) is nonmem.model.Model:
        model = input
        model = model.replace(name='start')
    else:
        # Redundant with validation
        raise TypeError(
            f'Invalid input: got `{input}` of type {type(input)},'
            f' only NONMEM model or standalone dataset are supported currently.'
        )

    # FIXME : Handle validation differently?
    # AMD start model (dataset) is required before validation
    args['input'] = model
    to_be_skipped = validate_input(**args)

    if parameter_uncertainty_method is not None:
        model = add_parameter_uncertainty_step(model, parameter_uncertainty_method)

    if lloq_method is not None:
        model = transform_blq(
            model,
            method=lloq_method,
            lloq=lloq_limit,
        )

    if strategy == "all":
        order = ['structural', 'iivsearch', 'residual', 'iovsearch', 'allometry', 'covariates']
    elif strategy == "retries":
        order = [
            'structural',
            'iivsearch',
            'residual',
            'iovsearch',
            'allometry',
            'covariates',
            'iivsearch',
            'residual',
        ]
    elif strategy == 'SIR':
        order = ['structural', 'iivsearch', 'residual']
    elif strategy == 'SRI':
        order = ['structural', 'residual', 'iivsearch']
    elif strategy == 'RSI':
        order = ['residual', 'structural', 'iivsearch']

    if modeltype == 'pkpd':
        warnings.warn('Skipping allometry since modeltype is "pkpd"')
        order.remove('allometry')

    if to_be_skipped:
        order = [tool for tool in order if tool not in to_be_skipped]

    try:
        input_search_space_features = [] if search_space is None else mfl_parse(search_space)
    except:  # noqa E722
        raise ValueError(f'Invalid `search_space`, could not be parsed: "{search_space}"')

    modelsearch_features = mfl_filtering(input_search_space_features, "modelsearch")
    if modeltype in ['pkpd', 'drug_metabolite']:
        structsearch_features = mfl_filtering(input_search_space_features, "structsearch")
        if search_space is None:
            if modeltype == 'pkpd':
                structsearch_features = "DIRECTEFFECT(*);EFFECTCOMP(*);INDIRECTEFFECT(*,*)"
            else:
                if administration in ['oral', 'ivoral']:
                    structsearch_features = "METABOLITE([PSC, BASIC]);PERIPHERALS([0,1], MET)"
                else:
                    structsearch_features = "METABOLITE([BASIC]);PERIPHERALS([0,1], MET)"

    if not modelsearch_features:
        if modeltype in ('basic_pk', 'drug_metabolite') and administration == 'oral':
            modelsearch_features = (
                Absorption((Name('FO'), Name('ZO'), Name('SEQ-ZO-FO'))),
                Elimination((Name('FO'),)),
                LagTime((Name('OFF'), Name('ON'))),
                Transits((0, 1, 3, 10), Wildcard()),
                Peripherals((0, 1)),
            )
        elif modeltype in ('basic_pk', 'drug_metabolite') and administration == 'ivoral':
            modelsearch_features = (
                Absorption((Name('FO'), Name('ZO'), Name('SEQ-ZO-FO'))),
                Elimination((Name('FO'),)),
                LagTime((Name('OFF'), Name('ON'))),
                Transits((0, 1, 3, 10), Wildcard()),
                Peripherals((0, 1, 2)),
            )
        elif modeltype == 'tmdd' and administration == 'oral':
            modelsearch_features = (
                Absorption((Name('FO'), Name('ZO'), Name('SEQ-ZO-FO'))),
                Elimination((Name('MM'), Name('MIX-FO-MM'))),
                LagTime((Name('OFF'), Name('ON'))),
                Transits((0, 1, 3, 10), Wildcard()),
                Peripherals((0, 1)),
            )
        elif modeltype == 'tmdd' and administration == 'ivoral':
            modelsearch_features = (
                Absorption((Name('FO'), Name('ZO'), Name('SEQ-ZO-FO'))),
                Elimination((Name('MM'), Name('MIX-FO-MM'))),
                LagTime((Name('OFF'), Name('ON'))),
                Transits((0, 1, 3, 10), Wildcard()),
                Peripherals((0, 1, 2)),
            )
        else:
            modelsearch_features = (
                Elimination((Name('FO'),)),
                Peripherals((0, 1, 2)),
            )

    covsearch_features = tuple(
        filter(
            lambda statement: isinstance(statement, COVSEARCH_STATEMENT_TYPES),
            input_search_space_features,
        )
    )
    if not any(map(lambda statement: isinstance(statement, Covariate), covsearch_features)):
        def_cov_search_feature = (
            Covariate(Ref('IIV'), Ref('CONTINUOUS'), ('exp',), '*', Option(True)),
            Covariate(Ref('IIV'), Ref('CATEGORICAL'), ('cat',), '*', Option(True)),
        )
        if modeltype == 'basic_pk' and administration == 'ivoral':
            def_cov_search_feature = def_cov_search_feature + (
                Covariate(('RUV',), ('ADMID',), ('cat',), '*', Option(True)),
            )

        covsearch_features = def_cov_search_feature + covsearch_features

    if modeltype == "tmdd":
        orig_dataset = model.dataset
        if dv_types is not None:
            model = filter_dataset(model, 'DVID < 2')

    db = default_tool_database(toolname='amd', path=path, exist_ok=resume)
    run_subfuncs = {}

    for section in order:
        if section == 'structural':
            run_subfuncs['structural_covariates'] = _subfunc_structural_covariates(
                amd_start_model=model,
                search_space=covsearch_features,
                strictness=strictness,
                path=db.path,
            )
            if modeltype == 'pkpd':
                func = _subfunc_structsearch(
                    type=modeltype,
                    search_space=structsearch_features,
                    b_init=b_init,
                    emax_init=emax_init,
                    ec50_init=ec50_init,
                    met_init=met_init,
                    strictness=strictness,
                    path=db.path,
                )
                run_subfuncs['structsearch'] = func
            elif modeltype == 'tmdd':
                func = _subfunc_structsearch_tmdd(
                    search_space=modelsearch_features,
                    type=modeltype,
                    strictness=strictness,
                    dv_types=dv_types,
                    orig_dataset=orig_dataset,
                    path=db.path,
                )
                run_subfuncs['structsearch'] = func
            else:
                func = _subfunc_modelsearch(
                    search_space=modelsearch_features, strictness=strictness, path=db.path
                )
                run_subfuncs['modelsearch'] = func
            # Perfomed 'after' modelsearch
            if modeltype == 'drug_metabolite':
                func = _subfunc_structsearch(
                    type=modeltype,
                    search_space=structsearch_features,
                    path=db.path,
                )
                run_subfuncs['structsearch'] = func
        elif section == 'iivsearch':
            if 'iivsearch' in run_subfuncs.keys():
                run_name = 'rerun_iivsearch'
                func = _subfunc_iiv(
                    iiv_strategy='no_add',
                    strictness=strictness,
                    path=db.path,
                    dir_name="rerun_iivsearch",
                )
            else:
                run_name = 'iivsearch'
                func = _subfunc_iiv(
                    iiv_strategy=iiv_strategy,
                    strictness=strictness,
                    path=db.path,
                    dir_name="iivsearch",
                )
            run_subfuncs[run_name] = func
        elif section == 'iovsearch':
            func = _subfunc_iov(
                amd_start_model=model, occasion=occasion, strictness=strictness, path=db.path
            )
            run_subfuncs['iovsearch'] = func
        elif section == 'residual':
            if any(k.startswith('ruvsearch') for k in run_subfuncs.keys()):
                run_name = 'rerun_ruvsearch'
            else:
                run_name = 'ruvsearch'
            if modeltype == 'drug_metabolite':
                # FIXME : Assume the dv number?
                # Perform two searches
                # One for the drug
                func = _subfunc_ruvsearch(
                    dv=1,
                    strictness=strictness,
                    path=db.path / f'{run_name}_drug',
                    dir_name='{run_name}_drug',
                )
                run_subfuncs[f'{run_name}_drug'] = func
                # And one for the metabolite
                func = _subfunc_ruvsearch(
                    dv=2,
                    strictness=strictness,
                    path=db.path / f'{run_name}_metabolite',
                    dir_name='{run_name}_metabolite',
                )
                run_subfuncs[f'{run_name}_metabolite'] = func
            elif modeltype == 'tmdd' and dv_types is not None:
                for key, value in dv_types.items():
                    func = _subfunc_ruvsearch(
                        dv=value, strictness=strictness, path=db.path / f'ruvsearch_{key}'
                    )
                    run_subfuncs[f'ruvsearch_{key}'] = func
            else:
                func = _subfunc_ruvsearch(
                    dv=dv, strictness=strictness, path=db.path, dir_name=run_name
                )
                run_subfuncs[f'{run_name}'] = func
        elif section == 'allometry':
            func = _subfunc_allometry(
                amd_start_model=model, allometric_variable=allometric_variable, path=db.path
            )
            run_subfuncs['allometry'] = func
        elif section == 'covariates':
            func = _subfunc_mechanistic_exploratory_covariates(
                amd_start_model=model,
                search_space=covsearch_features,
                mechanistic_covariates=mechanistic_covariates,
                strictness=strictness,
                path=db.path,
            )
            run_subfuncs['covsearch'] = func
        else:
            raise ValueError(f"Unrecognized section {section} in order.")
        if retries_strategy == 'all_final':
            func = _subfunc_retires(tool=section, strictness=strictness, seed=seed, path=db.path)
            run_subfuncs[f'{section}_retries'] = func

    if retries_strategy == 'final':
        func = _subfunc_retires(tool="", strictness=strictness, seed=seed, path=db.path)
        run_subfuncs['retries'] = func

    # Filter data to only contain dvid=1
    if modeltype == "drug_metabolite":
        orig_dataset = model.dataset
        # FIXME : remove
        model = filter_dataset(model, 'DVID != 2')

    if results is None:
        results = run_tool('modelfit', model, path=db.path / 'modelfit', resume=resume)
    model = model.replace(modelfit_results=results)
    next_model = model
    sum_subtools, sum_models, sum_inds_counts, sum_amd = [], [], [], []
    sum_subtools.append(_create_sum_subtool('start', model))
    for tool_name, func in run_subfuncs.items():
        if modeltype == 'drug_metabolite' and tool_name == "structsearch":
            next_model = next_model.replace(dataset=orig_dataset)
        subresults = func(next_model)

        if subresults is None:
            sum_models.append(None)
            sum_inds_counts.append(None)
        else:
            if subresults.final_model.name != next_model.name:
                if tool_name == "allometry" and 'allometry' in order[: order.index('covariates')]:
                    cov_before = ModelFeatures.create_from_mfl_string(
                        get_model_features(next_model)
                    )
                    cov_after = ModelFeatures.create_from_mfl_string(
                        get_model_features(subresults.final_model)
                    )
                    cov_differences = (cov_after - cov_before).covariate
                    if cov_differences:
                        covsearch_features += cov_differences
                        func = _subfunc_mechanistic_exploratory_covariates(
                            amd_start_model=model,
                            search_space=covsearch_features,
                            strictness=strictness,
                            mechanistic_covariates=mechanistic_covariates,
                            path=db.path,
                        )
                        run_subfuncs['covsearch'] = func
                next_model = subresults.final_model
                results = subresults.tool_database.model_database.retrieve_modelfit_results(
                    subresults.final_model.name
                )
                next_model = next_model.replace(modelfit_results=results)
            sum_subtools.append(_create_sum_subtool(tool_name, next_model))
            sum_models.append(subresults.summary_models.reset_index())
            sum_inds_counts.append(subresults.summary_individuals_count.reset_index())

    for sums in [sum_models, sum_inds_counts]:
        filtered_results = list(
            zip(*filter(lambda t: t[1] is not None, zip(list(run_subfuncs.keys()), sums)))
        )

        if not filtered_results:
            sum_amd.append(None)
            continue

        sums = pd.concat(
            filtered_results[1], keys=list(filtered_results[0]), names=['tool', 'default index']
        ).reset_index()
        if 'step' in sums.columns:
            sums['step'] = sums['step'].fillna(1).astype('int64')
        else:
            sums['step'] = 1

        sums.set_index(['tool', 'step', 'model'], inplace=True)
        sums.drop('default index', axis=1, inplace=True)
        sum_amd.append(sums)

    summary_models, summary_individuals_count = sum_amd
    summary_tool = _create_tool_summary(sum_subtools)

    if summary_models is None:
        warnings.warn(
            'AMDResults.summary_models is None because none of the tools yielded a summary.'
        )

    if summary_individuals_count is None:
        warnings.warn(
            'AMDResults.summary_individuals_count is None because none of the tools yielded '
            'a summary.'
        )

    final_model = next_model
    final_results = next_model.modelfit_results
    summary_errors = summarize_errors(final_results)
    if final_results.predictions is not None:
        dv_vs_ipred_plot = plot_dv_vs_ipred(model, final_results.predictions)
        dv_vs_pred_plot = plot_dv_vs_pred(model, final_results.predictions)
    else:
        dv_vs_ipred_plot = None
        dv_vs_pred_plot = None
    if final_results.residuals is not None:
        cwres_vs_idv_plot = plot_cwres_vs_idv(model, final_results.residuals)
    else:
        cwres_vs_idv_plot = None
    if final_results.predictions is not None and final_results.residuals is not None:
        abs_cwres_vs_ipred_plot = plot_abs_cwres_vs_ipred(
            model, predictions=final_results.predictions, residuals=final_results.residuals
        )
    else:
        abs_cwres_vs_ipred_plot = None
    res = AMDResults(
        final_model=final_model.name,
        summary_tool=summary_tool,
        summary_models=summary_models,
        summary_individuals_count=summary_individuals_count,
        summary_errors=summary_errors,
        final_model_parameter_estimates=_table_final_parameter_estimates(
            model, final_results.parameter_estimates_sdcorr, final_results.standard_errors_sdcorr
        ),
        final_model_dv_vs_ipred_plot=dv_vs_ipred_plot,
        final_model_dv_vs_pred_plot=dv_vs_pred_plot,
        final_model_cwres_vs_idv_plot=cwres_vs_idv_plot,
        final_model_abs_cwres_vs_ipred_plot=abs_cwres_vs_ipred_plot,
    )
    # Since we are outside of the regular tools machinery the following is needed
    results_path = db.path / 'results.json'
    write_results(results=res, path=results_path)
    write_results(results=res, path=db.path / 'results.csv', csv=True)
    rst_path = Path(__file__).parent / 'report.rst'
    target_path = db.path / 'results.html'
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        generate_report(rst_path, results_path, target_path)
    return res


def _table_final_parameter_estimates(model: Model, parameter_estimates, ses):
    rse = ses / parameter_estimates
    rse.name = "RSE"
    df = pd.concat([parameter_estimates, rse], axis=1)
    return df


def _create_sum_subtool(tool_name, selected_model):
    return {
        'tool': tool_name,
        'selected_model': selected_model.name,
        'description': selected_model.description,
        'n_params': len(selected_model.parameters.nonfixed),
        'ofv': selected_model.modelfit_results.ofv,
    }


def _create_tool_summary(rows):
    summary_prev = None
    rows_updated = []
    for summary in rows:
        summary_updated = summary
        if not summary_prev:
            summary_updated['d_params'] = 0
            summary_updated['dofv'] = 0
        else:
            summary_updated['d_params'] = summary['n_params'] - summary_prev['n_params']
            summary_updated['dofv'] = summary_prev['ofv'] - summary['ofv']
        rows_updated.append(summary_updated)
        summary_prev = summary

    columns = ['tool', 'selected_model', 'description', 'ofv', 'dofv', 'n_params', 'd_params']
    df = pd.DataFrame.from_records(rows_updated, columns=columns).set_index(['tool'])
    return df


SubFunc = Callable[[Model], Optional[Results]]


def noop_subfunc(_: Model):
    return None


def _subfunc_retires(tool, strictness, seed, path):
    def _run_retries(model):
        res = run_tool(
            'retries',
            model=model,
            results=model.modelfit_results,
            strictness=strictness,
            scale='UCP',
            prefix_name=tool,
            seed=seed,
            path=path / f'{tool}_retries',
        )
        assert isinstance(res, Results)
        return res

    return _run_retries


def _subfunc_modelsearch(search_space: Tuple[Statement, ...], strictness, path) -> SubFunc:
    def _run_modelsearch(model):
        res = run_tool(
            'modelsearch',
            search_space=mfl_stringify(search_space),
            algorithm='reduced_stepwise',
            model=model,
            strictness=strictness,
            results=model.modelfit_results,
            path=path / 'modelsearch',
        )
        assert isinstance(res, Results)

        return res

    return _run_modelsearch


def _subfunc_structsearch(path, **kwargs) -> SubFunc:
    def _run_structsearch(model):
        res = run_tool(
            'structsearch',
            model=model,
            results=model.modelfit_results,
            **kwargs,
            path=path / 'structsearch',
        )
        assert isinstance(res, Results)
        return res

    return _run_structsearch


def _subfunc_structsearch_tmdd(
    search_space, type, strictness, dv_types, orig_dataset, path
) -> SubFunc:
    def _run_structsearch_tmdd(model):
        res = run_tool(
            'modelsearch',
            search_space=mfl_stringify(search_space),
            algorithm='reduced_stepwise',
            model=model,
            strictness=strictness,
            results=model.modelfit_results,
            path=path / 'modelsearch',
        )

        final_model = res.final_model

        if not has_mixed_mm_fo_elimination(final_model):
            # Only select models that have mixed MM FO elimination
            # If no model with mixed MM FO then final model from modelsearch will be used
            model_db = res.tool_database.model_database
            all_models = [model_db.retrieve_model(model) for model in model_db.list_models()]
            models_mixed_mm_fo_el = [
                model.name for model in all_models if has_mixed_mm_fo_elimination(model)
            ]
            if len(models_mixed_mm_fo_el) > 0:
                rank_all = res.summary_tool.dropna(subset='bic')[['rank']]
                rank_filtered = rank_all.query('model in @models_mixed_mm_fo_el')
                if len(rank_filtered) > 0:
                    rank_filtered = rank_filtered.sort_values(by=['rank'])
                    highest_ranked = rank_filtered.index[0]
                    final_model = retrieve_models(path / 'modelsearch', names=[highest_ranked])[0]

        extra_model = None
        extra_model_results = None
        n_peripherals = len(final_model.statements.ode_system.find_peripheral_compartments())
        modelfeatures = ModelFeatures.create_from_mfl_string(get_model_features(final_model))
        # Model features - 1 peripheral compartment
        modelfeatures_minus = modelfeatures.replace(peripherals=Peripherals((n_peripherals - 1,)))
        # Loop through all models and find one with same features
        models = [
            model.name
            for model in all_models
            if ModelFeatures.create_from_mfl_string(get_model_features(model))
            == modelfeatures_minus
        ]
        if len(models) > 0:
            # Find highest ranked model
            rank_all = res.summary_tool.dropna(subset='bic')[['rank']]
            rank_filtered = rank_all.query('model in @models')
            if len(rank_filtered) > 0:
                rank_filtered = rank_filtered.sort_values(by=['rank'])
                highest_ranked = rank_filtered.index[0]
                extra_model = retrieve_models(path / 'modelsearch', names=[highest_ranked])[0]
                if dv_types is not None:
                    extra_model = extra_model.replace(dataset=orig_dataset)
                extra_model_results = extra_model.modelfit_results

        # Replace original dataset if multiple DVs
        if dv_types is not None:
            final_model = final_model.replace(dataset=orig_dataset)

        res = run_tool(
            'structsearch',
            type=type,
            model=final_model,
            results=final_model.modelfit_results,
            extra_model=extra_model,
            extra_model_results=extra_model_results,
            strictness=strictness,
            dv_types=dv_types,
            path=path / 'structsearch',
        )
        assert isinstance(res, Results)
        return res

    return _run_structsearch_tmdd


def _subfunc_iiv(iiv_strategy, strictness, path, dir_name) -> SubFunc:
    def _run_iiv(model):
        res = run_tool(
            'iivsearch',
            'brute_force',
            iiv_strategy=iiv_strategy,
            model=model,
            results=model.modelfit_results,
            strictness=strictness,
            keep=['CL'],
            path=path / dir_name,
        )
        assert isinstance(res, Results)
        return res

    return _run_iiv


def _subfunc_ruvsearch(dv, strictness, path, dir_name) -> SubFunc:
    def _run_ruvsearch(model):
        if has_blq_transformation(model):
            skip, max_iter = ['IIV_on_RUV', 'time_varying'], 1
        else:
            skip, max_iter = [], 3
        res = run_tool(
            'ruvsearch',
            model,
            results=model.modelfit_results,
            skip=skip,
            max_iter=max_iter,
            dv=dv,
            strictness=strictness,
            path=path / dir_name,
        )
        assert isinstance(res, Results)
        return res

    return _run_ruvsearch


def _subfunc_structural_covariates(
    amd_start_model: Model,
    search_space: Tuple[Statement, ...],
    strictness,
    path,
) -> SubFunc:
    def _run_structural_covariates(model):
        from pharmpy.modeling import get_pk_parameters

        allowed_parameters = allowed_parameters = set(get_pk_parameters(model)).union(
            str(statement.symbol) for statement in model.statements.before_odes
        )
        # Extract all forced
        mfl = ModelFeatures.create_from_mfl_statement_list(search_space)
        mfl_covariates = mfl.expand(model).covariate
        structural_searchspace = []
        skipped_parameters = set()
        for cov_statement in mfl_covariates:
            if not cov_statement.optional.option:
                filtered_parameters = tuple(
                    [p for p in cov_statement.parameter if p in allowed_parameters]
                )
                # Not optional -> Add to all search spaces (was added in structural run)
                structural_searchspace.append(
                    Covariate(
                        filtered_parameters,
                        cov_statement.covariate,
                        cov_statement.fp,
                        cov_statement.op,
                        cov_statement.optional,
                    )
                )
                skipped_parameters.union(set(cov_statement.parameter) - set(filtered_parameters))

        # Ignore warning?
        if skipped_parameters:
            warnings.warn(
                f'{skipped_parameters} missing in start model and structural covariate effect cannot be added'
                ' Might be added during a later COVsearch step if possible.'
            )
        if not structural_searchspace and not skipped_parameters:
            # Uneccessary to warn (?)
            return None
        elif not structural_searchspace:
            warnings.warn(
                'No applicable structural covariates found in search space. Skipping structural_COVsearch'
            )
            return None

        res = run_tool(
            'covsearch',
            mfl_stringify(structural_searchspace),
            model=model,
            strictness=strictness,
            results=model.modelfit_results,
            path=path / 'covsearch_structural',
        )
        assert isinstance(res, Results)
        return res

    return _run_structural_covariates


def _subfunc_mechanistic_exploratory_covariates(
    amd_start_model: Model,
    search_space: Tuple[Statement, ...],
    mechanistic_covariates,
    strictness,
    path,
) -> SubFunc:
    covariates = set(extract_covariates(amd_start_model, search_space))
    if covariates:
        allowed_covariates = get_covariates_allowed_in_covariate_effect(amd_start_model)
        for covariate in sorted(covariates):
            if covariate not in allowed_covariates:
                raise ValueError(
                    f'Invalid `search_space` because of invalid covariate found in'
                    f' search_space: got `{covariate}`,'
                    f' must be in {sorted(allowed_covariates)}.'
                )
    else:
        warnings.warn(
            'COVsearch will most likely be skipped because no covariates could be found.'
            ' Check search_space definition'
            ' and .datainfo usage of "covariate" type and "continuous" flag.'
        )

    def _run_mechanistic_exploratory_covariates(model):
        index_offset = 0  # For naming runs

        effects = list(covariate_spec(model, search_space))

        if not effects:
            warnings.warn(
                'Skipping COVsearch because no effect candidates could be generated.'
                ' Check search_space definition'
                ' and .datainfo usage of "covariate" type and "continuous" flag.'
            )
            return None

        if mechanistic_covariates:
            # Extract them and all forced
            mfl = ModelFeatures.create_from_mfl_statement_list(search_space)
            mfl_covariates = mfl.expand(model).covariate
            filtered_searchspace = []
            mechanistic_searchspace = []
            structural_searchspace = []
            for cov_statement in mfl_covariates:
                if not cov_statement.optional.option:
                    # Not optional -> Add to all search spaces (was added in structural run)
                    structural_searchspace.append(cov_statement)
                    mechanistic_searchspace.append(cov_statement)
                    filtered_searchspace.append(cov_statement)
                else:
                    current_cov = []
                    for cov in cov_statement.covariate:
                        if cov in mechanistic_covariates:
                            current_cov.append(cov)
                    if current_cov:
                        mechanistic_cov = Covariate(
                            cov_statement.parameter,
                            tuple(current_cov),
                            cov_statement.fp,
                            cov_statement.op,
                            cov_statement.optional,
                        )
                        mechanistic_searchspace.append(mechanistic_cov)
                        if filtered_cov_set := tuple(
                            set(cov_statement.covariate) - set(current_cov)
                        ):
                            filtered_cov = Covariate(
                                cov_statement.parameter,
                                filtered_cov_set,
                                cov_statement.fp,
                                cov_statement.op,
                                cov_statement.optional,
                            )
                            filtered_searchspace.append(filtered_cov)
                    else:
                        filtered_searchspace.append(cov_statement)

            if not mechanistic_searchspace:
                warnings.warn(
                    'No covariate effect for given mechanistic covariates found.'
                    ' Skipping mechanistic COVsearch.'
                )
            else:
                res = run_tool(
                    'covsearch',
                    mfl_stringify(mechanistic_searchspace),
                    model=model,
                    strictness=strictness,
                    results=model.modelfit_results,
                    path=path / 'covsearch_mechanistic',
                )
                index_offset = int(
                    res.summary_tool.iloc[-1].name[-1][13:]
                )  # Get largest number of run
                if res.final_model.name != model.name:
                    model = res.final_model
                    model_results = res.tool_database.model_database.retrieve_modelfit_results(
                        res.final_model.name
                    )
                    model = model.replace(modelfit_results=model_results)
                    added_covs = ModelFeatures.create_from_mfl_string(
                        get_model_features(model)
                    ).covariate
                    filtered_searchspace.extend(
                        added_covs
                    )  # Avoid removing added cov in exploratory
        else:
            filtered_searchspace = search_space

        res = run_tool(
            'covsearch',
            mfl_stringify(filtered_searchspace),
            model=model,
            strictness=strictness,
            results=model.modelfit_results,
            path=path / 'covsearch_exploratory',
            naming_index_offset=index_offset,
        )
        assert isinstance(res, Results)
        return res

    return _run_mechanistic_exploratory_covariates


def _subfunc_allometry(amd_start_model: Model, allometric_variable, path) -> SubFunc:
    if allometric_variable is None:  # Somewhat redundant with validation function
        allometric_variable = amd_start_model.datainfo.descriptorix["body weight"][0].name

    def _run_allometry(model):
        res = run_tool(
            'allometry',
            model,
            results=model.modelfit_results,
            allometric_variable=allometric_variable,
            path=path / 'allometry',
        )
        assert isinstance(res, Results)
        return res

    return _run_allometry


def _subfunc_iov(amd_start_model, occasion, strictness, path) -> SubFunc:
    def _run_iov(model):
        res = run_tool(
            'iovsearch',
            model=model,
            results=model.modelfit_results,
            column=occasion,
            strictness=strictness,
            path=path / 'iovsearch',
        )
        assert isinstance(res, Results)
        return res

    return _run_iov


@with_runtime_arguments_type_check
def validate_input(
    input: Model,
    results: Optional[ModelfitResults] = None,
    modeltype: str = 'basic_pk',
    administration: str = 'oral',
    strategy: str = "all",
    cl_init: float = 0.01,
    vc_init: float = 1.0,
    mat_init: float = 0.1,
    b_init: Optional[Union[int, float]] = None,
    emax_init: Optional[Union[int, float]] = None,
    ec50_init: Optional[Union[int, float]] = None,
    met_init: Optional[Union[int, float]] = None,
    search_space: Optional[str] = None,
    lloq_method: Optional[str] = None,
    lloq_limit: Optional[str] = None,
    allometric_variable: Optional[Union[str, sympy.Symbol]] = None,
    occasion: Optional[str] = None,
    path: Optional[Union[str, Path]] = None,
    resume: bool = False,
    strictness: Optional[str] = "minimization_successful or (rounding_errors and sigdigs>=0.1)",
    dv_types: Optional[dict[Literal[DV_TYPES], int]] = None,
    mechanistic_covariates: Optional[List[str]] = None,
    retries_strategy: Literal["final", "all_final", "skip"] = "all_final",
    seed: Optional[Union[np.random.Generator, int]] = None,
    parameter_uncertainty_method: Optional[Literal['SANDWICH', 'CPG', 'OFIM']] = None,
    ignore_datainfo_fallback: bool = False,
):
    model = input
    to_be_skipped = []

    check_list("modeltype", modeltype, ALLOWED_MODELTYPE)

    check_list("administration", administration, ALLOWED_ADMINISTRATION)

    check_list("strategy", strategy, ALLOWED_STRATEGY)

    if search_space is not None:
        try:
            input_search_space_features = [] if search_space is None else mfl_parse(search_space)
        except:  # noqa E722
            raise ValueError(f'Invalid `search_space`, could not be parsed: "{search_space}"')

    check_list("retries_strategy", retries_strategy, RETRIES_STRATEGIES)

    # IOVSEARCH
    if occasion is None:
        warnings.warn('IOVsearch will be skipped because occasion is None.')
        to_be_skipped.append("iovsearch")
    else:
        if occasion not in model.dataset:
            raise ValueError(
                f'Invalid `occasion`: got `{occasion}`,'
                f' must be one of {sorted(model.datainfo.names)}.'
            )
        categories = get_occasion_levels(model.dataset, occasion)
        if len(categories) < 2:
            warnings.warn(
                f'Skipping IOVsearch because there are less than two '
                f'occasion categories in column "{occasion}": {categories}.'
            )
            to_be_skipped.append("iovsearch")

    # ALLOMETRY
    if allometric_variable is None:
        if not ignore_datainfo_fallback:
            try:
                model.datainfo.descriptorix["body weight"]
            except IndexError:
                warnings.warn(
                    'Allometry will be skipped because allometric_variable is None and could'
                    ' not be inferred through .datainfo via "body weight" descriptor.'
                )
                to_be_skipped.append("allometry")
        else:
            warnings.warn(
                'Allometry will be skipped because allometric_variable is None and'
                ' ignore_datainfo_fallback is True'
            )
            to_be_skipped.append("allometry")
    else:
        validate_allometric_variable(model, allometric_variable)

    # COVSEARCH
    if search_space is not None:
        covsearch_features = tuple(
            filter(
                lambda statement: isinstance(statement, COVSEARCH_STATEMENT_TYPES),
                input_search_space_features,
            )
        )
        if covsearch_features:  # Check LET() and COVARIATE()
            covariates = set(extract_covariates(model, input_search_space_features))
            let_features = [
                statement for statement in covsearch_features if isinstance(statement, Let)
            ]
            for statement in let_features:
                covariates = covariates.union(set(statement.value))
            if covariates:
                allowed_covariates = get_covariates_allowed_in_covariate_effect(model)
                for covariate in sorted(covariates):
                    if covariate not in allowed_covariates:
                        raise ValueError(
                            f'Invalid `search_space` because of invalid covariate found in'
                            f' search_space: got `{covariate}`,'
                            f' must be in {sorted(allowed_covariates)}.'
                        )
            else:
                warnings.warn(
                    'COVsearch will be skipped because no covariates could be found'
                    ' in the given search space.'
                    ' Check search_space definition'
                )
                to_be_skipped.append("covariates")
        else:
            if ignore_datainfo_fallback:
                warnings.warn(
                    'COVsearch will be skipped because no covariates were given'
                    ' and ignore_datainfo_fallback is True.'
                )
                to_be_skipped.append("covariates")
            elif not any(column.type == 'covariate' for column in model.datainfo):
                warnings.warn(
                    'COVsearch will be skipped because no covariates were given'
                    ' or could be extracted.'
                    ' Check search_space definition'
                    ' and .datainfo usage of "covariate" type and "continuous" flag.'
                )
                to_be_skipped.append("covariates")
    else:
        if ignore_datainfo_fallback:
            warnings.warn(
                'COVsearch will be skipped because no covariates were given'
                ' and ignore_datainfo_fallback is True.'
            )
            to_be_skipped.append("covariates")
        elif not any(column.type == 'covariate' for column in model.datainfo):
            warnings.warn(
                'COVsearch will be skipped because no covariates were given'
                ' or could be extracted.'
                ' Check search_space definition'
                ' and .datainfo usage of "covariate" type and "continuous" flag.'
            )
            to_be_skipped.append("covariates")

    return to_be_skipped



import pytest

from pharmpy.deps import numpy as np
from pharmpy.model.model import update_datainfo
from pharmpy.modeling import (
    create_joint_distribution,
    remove_error_model,
    set_additive_error_model,
    set_combined_error_model,
    set_iiv_on_ruv,
    set_power_on_ruv,
    set_proportional_error_model,
    transform_blq,
)
from pharmpy.modeling.blq import has_blq_transformation


@pytest.mark.parametrize(
    'method, error_func, sd_ref, y_ref',
    [
        (
            'm4',
            set_additive_error_model,
            'SD = SQRT(SIGMA(1,1))',
            ('Y = F + EPS(1)', 'Y = (CUMD - CUMDZ)/(1 - CUMDZ)'),
        ),
        (
            'm4',
            set_proportional_error_model,
            'SD = SQRT(SIGMA(1,1))*ABS(F)',
            ('Y = F + EPS(1)*F', 'Y = (CUMD - CUMDZ)/(1 - CUMDZ)'),
        ),
        (
            'm4',
            set_combined_error_model,
            'SD = SQRT(F**2*SIGMA(1,1) + SIGMA(2,2))',
            ('Y = F + EPS(1)*F + EPS(2)', 'Y = (CUMD - CUMDZ)/(1 - CUMDZ)'),
        ),
        (
            'm4',
            set_power_on_ruv,
            'SD = SQRT(SIGMA(1,1))*SQRT(F**(2*THETA(3)))',
            ('Y = F + EPS(1)*F**THETA(3)', 'Y = (CUMD - CUMDZ)/(1 - CUMDZ)'),
        ),
        (
            'm3',
            set_additive_error_model,
            'SD = SQRT(SIGMA(1,1))',
            ('Y = F + EPS(1)', 'Y = PHI((-F + LLOQ)/SD)'),
        ),
    ],
)
def test_transform_blq(load_model_for_test, testdata, method, error_func, sd_ref, y_ref):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = error_func(model)

    model = transform_blq(model, method=method, lloq=0.1)

    assert sd_ref in model.model_code
    assert all(statement in model.model_code for statement in y_ref)

    assert all(est.laplace for est in model.estimation_steps)


@pytest.mark.parametrize(
    'method, error_func_before, error_func_after, args, sd_ref, y_ref',
    [
        (
            'm4',
            set_additive_error_model,
            set_proportional_error_model,
            {'zero_protection': False},
            'SD = SQRT(SIGMA(1,1))*ABS(F)',
            ('Y = F + EPS(1)*F', 'Y = (CUMD - CUMDZ)/(1 - CUMDZ)'),
        ),
        (
            'm4',
            set_additive_error_model,
            set_proportional_error_model,
            {},
            'SD = SQRT(SIGMA(1,1))*ABS(IPREDADJ)',
            ('Y = F + EPS(1)*IPREDADJ', 'Y = (CUMD - CUMDZ)/(1 - CUMDZ)'),
        ),
        (
            'm4',
            set_additive_error_model,
            set_combined_error_model,
            {},
            'SD = SQRT(F**2*SIGMA(1,1) + SIGMA(2,2))',
            ('Y = F + EPS(1)*F + EPS(2)', 'Y = (CUMD - CUMDZ)/(1 - CUMDZ)'),
        ),
        (
            'm4',
            set_proportional_error_model,
            set_combined_error_model,
            {},
            'SD = SQRT(F**2*SIGMA(1,1) + SIGMA(2,2))',
            ('Y = F + EPS(1)*F + EPS(2)', 'Y = (CUMD - CUMDZ)/(1 - CUMDZ)'),
        ),
        (
            'm4',
            set_additive_error_model,
            set_power_on_ruv,
            {},
            'SD = SQRT(F**(2*THETA(3))*SIGMA(1,1))',
            ('Y = F + EPS(1)*F**THETA(3)', 'Y = (CUMD - CUMDZ)/(1 - CUMDZ)'),
        ),
        (
            'm4',
            set_proportional_error_model,
            set_power_on_ruv,
            {},
            'SD = SQRT(IPREDADJ**(2*THETA(3))*SIGMA(1,1))',
            ('Y = F + EPS(1)*IPREDADJ**THETA(3)', 'Y = (CUMD - CUMDZ)/(1 - CUMDZ)'),
        ),
    ],
)
def test_update_blq_transformation(
    load_model_for_test, testdata, method, error_func_before, error_func_after, args, sd_ref, y_ref
):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = remove_error_model(model)
    model = error_func_before(model)

    model = transform_blq(model, method=method, lloq=0.1)

    model = error_func_after(model, **args)

    assert sd_ref in model.model_code
    assert all(statement in model.model_code for statement in y_ref)

    assert all(est.laplace for est in model.estimation_steps)


@pytest.mark.parametrize(
    'method, error_func',
    [
        ('m4', set_additive_error_model),
        ('m4', set_proportional_error_model),
        ('m4', set_combined_error_model),
        ('m4', set_power_on_ruv),
        ('m3', set_additive_error_model),
    ],
)
def test_has_blq_transformation(load_model_for_test, testdata, method, error_func):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = error_func(model)

    assert not has_blq_transformation(model)

    model = transform_blq(model, method=method, lloq=0.1)

    assert has_blq_transformation(model)


def test_transform_blq_invalid_input_model(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = set_combined_error_model(model)
    model = create_joint_distribution(model, model.random_variables.epsilons.names)
    with pytest.raises(ValueError, match='Invalid input model: covariance between epsilons'):
        transform_blq(model, method='m4', lloq=0.1)

    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = set_iiv_on_ruv(model)
    with pytest.raises(ValueError, match='Invalid input model: error model not supported'):
        transform_blq(model, method='m4', lloq=0.1)


def test_transform_blq_different_lloq(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model_float = transform_blq(model, lloq=0.1)

    assert 'DV.GE.LLOQ' in model_float.model_code

    df_blq = model.dataset
    df_blq['BLQ'] = np.random.randint(0, 2, df_blq.shape[0])
    di_blq = update_datainfo(model.datainfo, df_blq)
    blq_col = di_blq['BLQ'].replace(type='blq')
    di_blq = di_blq.set_column(blq_col)
    model_blq = model.replace(dataset=df_blq, datainfo=di_blq)

    with pytest.raises(ValueError):
        transform_blq(model_blq)

    df_lloq = model.dataset
    df_lloq['LLOQ'] = np.random.random(df_lloq.shape[0])
    di_lloq = update_datainfo(model.datainfo, df_lloq)
    lloq_col = di_lloq['LLOQ'].replace(type='lloq')
    di_lloq = di_lloq.set_column(lloq_col)
    model_lloq = model.replace(dataset=df_lloq, datainfo=di_lloq)

    model_lloq_col = transform_blq(model_lloq)

    assert 'DV.GE.LLOQ' in model_lloq_col.model_code
    assert 'LLOQ = ' not in model_lloq_col.model_code

    model_float_with_blq_col = transform_blq(model_blq, lloq=0.1)

    assert 'BLQ.EQ.0' in model_float_with_blq_col.model_code
    assert 'LLOQ = ' in model_float_with_blq_col.model_code



from __future__ import annotations

from typing import TYPE_CHECKING, Set

from .leaves import free_images

if TYPE_CHECKING:
    import sympy
else:
    from pharmpy.deps import sympy


def canonical_ode_rhs(expr: sympy.Expr):
    fi = free_images(expr)
    return sympy.collect(_expand_rates(expr, fi), sorted(fi, key=str))


def _expand_rates(expr: sympy.Expr, free_images: Set[sympy.Expr]):
    if isinstance(expr, sympy.Add):
        return sympy.expand(
            sympy.Add(*map(lambda x: _expand_rates(x, free_images), expr.args)), deep=False
        )
    if (
        isinstance(expr, sympy.Mul)
        and len(expr.args) == 2
        and not free_images.isdisjoint(expr.args)
    ):
        return sympy.expand(expr, deep=False)
    return expr
